"""
Implementation of the contour classification scheme proposed by Huron (1996) [1]

The classification scheme is also included in the FANTASTIC toolbox of
Müllensiefen (2009) [2]

<small>**Author**: Mark Gotham</small>

References
----------
1. Huron, D (2006). The Melodic Arch in Western Folksongs. *Computing in Musicology* 10.

2. Müllensiefen, D. (2009). Fantastic: Feature ANalysis Technology Accessing
        STatistics (In a Corpus): Technical Report v1.5

"""

__author__ = "Mark Gotham"

from amads.core.utils import sign


class HuronContour:
    """The contour classification scheme proposed by Huron (1996) [1]

    The classification scheme is also included in the FANTASTIC toolbox
    of Müllensiefen (2009) [2] (as Feature 19 `Huron Contour: h.contour`).

    Huron categorises melodies by identifying the start, mean, and end pitches
    and describing contour in terms of the two directions: start-mean, and mean-end.
    """

    def __init__(
        self, pitches: list[int], times: list[float], method: str = "amads"
    ):
        """Initialize with pitch and time values.

        Parameters
        ----------
        pitches : list[int]
            Pitch values in any numeric format (e.g., MIDI numbers).
        times : list[float]
            Onset times in any consistent, proportional scheme
            (e.g., seconds, quarter notes, etc.)

        Raises
        ------
        ValueError
            If the `times` and `pitches` parameters are not the same length.

        Examples
        --------
        >>> happy_birthday_pitches = [
        ...     60, 60, 62, 60, 65, 64, 60, 60, 62, 60, 67, 65,
        ...     60, 60, 72, 69, 65, 64, 62, 70, 69, 65, 67, 65
        ... ]
        >>> happy_birthday_times = [
        ...     0, 0.75, 1, 2, 3, 4, 6, 6.75, 7, 8, 9, 10,
        ...     12, 12.75, 13, 14, 15, 16, 17, 18, 18.75, 19, 20, 21
        ... ]
        >>> hc = HuronContour(
        ...     happy_birthday_pitches,
        ...     happy_birthday_times,
        ... )

        >>> hc.first_pitch
        60
        >>> hc.mean_pitch
        65
        >>> hc.last_pitch
        65
        >>> hc.first_to_mean
        5
        >>> hc.mean_to_last
        0
        >>> hc.contour_class
        'Ascending-Horizontal'

        References
        ----------
          1. Huron, D (2006). The Melodic Arch in Western Folksongs.
             *Computing in Musicology* 10.

          2. Müllensiefen, D. (2009). Fantastic: Feature ANalysis Technology Accessing
             STatistics (In a Corpus): Technical Report v1.5
        """
        if len(times) != len(pitches):
            raise ValueError(
                f"Times and pitches must have the same length, "
                f"got {len(times)} and {len(pitches)}"
            )

        self.times = times
        self.pitches = pitches
        self.first_pitch = pitches[0]
        self.last_pitch = pitches[-1]

        self.mean_pitch = None
        self.first_to_mean = None
        self.mean_to_last = None
        self.calculate_mean_attributes()

        self.contour_class = None
        self.class_label()

    def calculate_mean_attributes(self):
        """
        Calculate the mean and populate the remaining attributes.

        Note that the mean pitch is rounded to the nearest integer,
        and that this rounding happens before calculating comparisons.
        """
        self.mean_pitch = int(
            sum(x * y for x, y in zip(self.pitches, self.times))
            / sum(self.times)
        )

        self.first_to_mean = self.mean_pitch - self.first_pitch
        self.mean_to_last = self.last_pitch - self.mean_pitch

    def class_label(self):
        """Classify a contour into Huron's categories.

        This is based simply on the two directions from start to mean and mean to last.
        Huron proposes shorthands for some of these as follows:
        "Ascending-Descending" = "Convex",
        "Ascending-Horizontal" = None,
        "Ascending-Ascending": None,
        "Horizontal-Descending": None,
        "Horizontal-Horizontal": "Horizontal",
        "Horizontal-Ascending": None,
        "Descending-Descending": "Descending",
        "Descending-Ascending": "Concave"

        Where no shorthand is provided, this method return the longhand.

        Returns
        -------
        str
            String, exactly as reported in the FANTASTIC library.

        """

        direction_dict = {-1: "Descending", 0: "Horizontal", 1: "Ascending"}

        first_to_mean_sign = sign(self.first_to_mean)
        mean_to_last_sign = sign(self.mean_to_last)

        return_string = f"{direction_dict[first_to_mean_sign]}-{direction_dict[mean_to_last_sign]}"

        shorthand_dict = {
            "Ascending-Descending": "Convex",
            "Horizontal-Horizontal": "Horizontal",
            "Descending-Descending": "Descending",
            "Descending-Ascending": "Concave",
        }

        if return_string in shorthand_dict:
            self.contour_class = shorthand_dict[return_string]
        else:
            self.contour_class = return_string



"""
Basic tests of error raising in metrical grid module.
"""

__author__ = "Mark Gotham"


from fractions import Fraction

import pytest

from amads.harmony.consonance.consonance import approximate_fraction_consonance
from amads.time.meter import grid, profiles

test_list = [0, 0.75, 1.5, 2]


def test_tatum_bins():

    # First, test valid approach
    assert grid.get_tatum_from_priorities(test_list) == Fraction(1, 4)

    # Not a list
    with pytest.raises(ValueError):
        grid.get_tatum_from_priorities(test_list, pulse_priority_list=1)

    # List items not integers
    with pytest.raises(ValueError):
        grid.get_tatum_from_priorities(test_list, pulse_priority_list=[1.6])

    # List items negative integer
    with pytest.raises(ValueError):
        grid.get_tatum_from_priorities(test_list, pulse_priority_list=[-1])


def test_tatum_distance_threshold():
    with pytest.raises(ValueError):
        grid.get_tatum_from_priorities(test_list, distance_threshold=-1)


def test_tatum_proportion_threshold():
    with pytest.raises(ValueError):
        grid.get_tatum_from_priorities(test_list, proportion_threshold=3)


def test_BPSD():
    """Tests all measure-relative values of the BPSD dataset"""
    all_files = [
        "op002No1_01",
        "op002No2_01",
        "op002No3_01",
        "op007_01",
        "op010No1_01",
        "op010No2_01",
        "op010No3_01",
        "op013_01",
        "op014No1_01",
        "op014No2_01",
        "op022_01",
        "op026_01",
        "op027No1_01",
        "op027No2_01",
        "op028_01",
        "op031No1_01",
        "op031No2_01",
        "op031No3_01",
        "op049No1_01",
        "op049No2_01",
        "op053_01",
        "op054_01",
        "op057_01",
        "op078_01",
        "op079_01",
        "op081a_01",
        "op090_01",
        "op101_01",
        "op106_01",
        "op109_01",
    ]
    all_positions = []

    bpsd = profiles.BPSD()
    for file_name in all_files:
        these_keys = getattr(bpsd, file_name).keys()
        all_positions += list(these_keys)
        for k in these_keys:
            n, d = approximate_fraction_consonance(k)
            assert n < 200
            assert d < 600

    assert len(set(all_positions)) == len(set(bpsd.all.keys())) == 205



"""
Provides the `nnotes` function

Original doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=77
"""

from amads.core.basics import Note, Score


def nnotes(score: Score, merge_ties: bool = False) -> int:
    """
    Returns the number of notes in a musical score.

    Parameters
    ----------
    score : Score
        The musical score to analyze

    merge_ties : bool
        Count tied sequences of notes as a single note.


    Returns
    -------
    int
        The number of notes in the score

    Examples
    --------
    >>> from amads.music import example
    >>> from amads.io.readscore import read_score
    >>> import contextlib
    >>> # Load example score while suppressing output:
    >>> with contextlib.redirect_stdout(None):
    ...     score = read_score(example.fullpath("musicxml/ex3.xml"))
    >>> nnotes(score)
    2
    >>> nnotes(score, merge_ties=True)
    1
    """

    if merge_ties:
        score = score.merge_tied_notes()  # type: ignore
    total = 0
    for _ in score.find_all(Note):
        total += 1
    return total



"""
Pitch interval transition analysis.

Implements the Midi Toolbox `ivdist2` function

Original doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=64
"""

from typing import cast

from amads.core.basics import Note, Part, Score
from amads.core.distribution import Distribution
from amads.core.histogram import Histogram2D
from amads.pitch.pcdist1 import duraccent


def interval_distribution_2(
    score: Score,
    name: str = "Interval Transition Distribution",
    weighted: bool = True,
    miditoolbox_compatible: bool = True,
) -> Distribution:
    """
    Returns the 2nd-order interval distribution of a musical score.

    Currently, intervals greater than an octave will be ignored.

    Parameters
    ----------
    score : Score
        The musical score to analyze
    name : str
        A name for the distribution and plot title.
    weighted : bool, optional
        If True, the interval distribution is weighted by note durations
        in seconds that are modified according to Parncutt's durational
        accent model (1994), by default True.
    miditoolbox_compatible : bool
        miditoolbox_compatible introduces four changes to emulate `ivdist2`
        in Midi Toolbox: (1) avoid zero division by dividing counts
        by (total count plus (1e-12 time the number of bins), as
        opposed to simply skipping division when all bins are zero,
        (2) assume octave (but not direction) equivalence, so the
        intervals +1 and +13 update the same bin (as opposed to ignoring
        intervals larger than an octave), (3) a zero interval (unison) is
        inserted at the beginning of the sequence, (4) the weight is the
        sum of the modified durations of the second interval (as opposed to
        taking the sum of the modified durations of all three notes).

    Returns
    -------
    Distribution
        A 25x25 distribution where where (i,j) represents the normalized
        probabilities of transitioning from interval i to interval j.
        The bins are spaced at semitone distances with the first bin
        representing the downward octave and the last bin representing
        the upward octave. If the score is empty, the function returns
        a list with all elements set to zero.

    Raises
    ------
    ValueError
        If the score is not monophonic (e.g. contains chords)
    """
    if not score.ismonophonic():
        raise ValueError("Error: Score must be monophonic")

    score = cast(Score, score.merge_tied_notes())
    if weighted:
        score.convert_to_seconds()  # need seconds for duraccent function

    bin_centers = [float(i - 12) for i in range(25)]  # 25 bins from -12 to +12
    bin_boundaries = [i - 12 - 0.5 for i in range(26)]  # boundaries
    x_categories = [str(c) for c in bin_centers]
    y_categories = x_categories
    h = Histogram2D(bin_centers, bin_boundaries, "linear", True)
    for p in score.find_all(Part):
        part: Part = cast(Part, p)
        dur = 0.0  # (this value is never used)
        prev_iv = 0 if miditoolbox_compatible else None  # previous interval
        prev_pitch = None
        prev_dur = 0
        prev_prev_dur = 0
        prev_bin = None
        for n in part.find_all(Note):
            note: Note = cast(Note, n)
            if weighted:
                dur = duraccent(note)
            if prev_pitch is not None:
                iv = round(note.key_num - prev_pitch)
                if miditoolbox_compatible:
                    iv = (abs(iv) % 12) * ((iv > 0) - (iv < 0))
                if weighted:
                    dur = duraccent(note)
                    w = prev_dur + dur
                    if not miditoolbox_compatible:
                        w += prev_prev_dur
                    prev_bin = h.add_point_2d(prev_iv, iv, w, prev_bin)
                    prev_prev_dur = prev_dur
                else:
                    prev_bin = h.add_point_2d(prev_iv, iv, 1.0, prev_bin)
                prev_iv = None if prev_bin is None else iv
            prev_pitch = note.key_num
            prev_dur = dur
    if miditoolbox_compatible:
        total = sum(sum(bin) for bin in h.bins) + (
            len(h.bins) * len(h.bins) * 1e-12
        )
        h.bins = [[c / total for c in row] for row in h.bins]
    else:  # normalize normally
        h.normalize()

    return Distribution(
        name,
        h.bins,
        "interval_transition",
        [25, 25],
        x_categories,  # type: ignore
        "Interval (from)",
        y_categories,  # type: ignore
        "Interval (to)",
    )



"""
Tests for the `attractor_tempos` module.

Tests functionality for calculating salience.
"""

import numpy as np
import pytest
from matplotlib import pyplot as plt

from amads.time.meter.attractor_tempos import MetricalSalience, log_gaussian
from amads.time.meter.representations import PulseLengths


@pytest.fixture
def metrical_salience_instance():
    """
    Provides a sample set of pulse lengths for testing
    and initialises an instance of MetricalSalience with them.
    """
    pl = [4, 2, 1, 0.5]
    pls = PulseLengths(pulse_lengths=pl, cycle_length=4)
    pulse_array = pls.to_array()
    return MetricalSalience(pulse_array, quarter_bpm=100)


def test_symbolic(metrical_salience_instance):
    """Tests the initialization of the MetricalSalience class."""
    np.testing.assert_allclose(
        metrical_salience_instance.symbolic_pulses[0, :],
        np.array([4.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
    )


def test_absolute_pulse_lengths(metrical_salience_instance):
    np.testing.assert_allclose(
        metrical_salience_instance.absolute_pulses,
        np.array(
            [
                [2.4, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.2, 0.0, 0.0, 0.0, 1.2, 0.0, 0.0, 0.0],
                [0.6, 0.0, 0.6, 0.0, 0.6, 0.0, 0.6, 0.0],
                [0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3, 0.3],
            ]
        ),
    )


def test_get_salience_values(metrical_salience_instance):
    """Tests the `calculate_salience_values` method."""
    metrical_salience_instance.calculate_salience_values()
    assert metrical_salience_instance.salience_values.shape == (4, 8)


def test_get_cumulative_salience_values(metrical_salience_instance):
    """Tests the `calculate_cumulative_salience_values` method."""
    assert metrical_salience_instance.cumulative_salience_values is not None
    assert metrical_salience_instance.cumulative_salience_values.shape == (8,)
    np.testing.assert_allclose(
        metrical_salience_instance.cumulative_salience_values,
        np.array(
            [
                2.342383,
                0.604448,
                1.604448,
                0.604448,
                2.208897,
                0.604448,
                1.604448,
                0.604448,
            ]
        ),
        atol=1e-06,
    )


def test_plot(metrical_salience_instance):
    """Tests the plot method with symbolic, and then salience data, showing the difference."""
    fig = metrical_salience_instance.plot(symbolic_not_absolute=True)
    plt.close(fig)

    metrical_salience_instance.calculate_salience_values()
    fig = metrical_salience_instance.plot(symbolic_not_absolute=False)
    plt.close(fig)


def test_log_gaussian():
    """Tests the `log_gaussian` function."""
    assert np.isclose(log_gaussian(0.6), np.float64(1.0))
    assert np.isclose(log_gaussian(1.2), np.float64(0.604448254722616))
    assert np.isclose(log_gaussian(0.5), np.float64(0.96576814))
    assert np.isclose(log_gaussian(0.0), np.float64(0.0))


def test_log_gaussian_raises():
    """Tests the `log_gaussian` function raises errors in these cases."""
    with pytest.raises(ValueError):
        log_gaussian(0.6, mu=-1)
    with pytest.raises(ValueError):
        log_gaussian(0.6, sig=-100)



"""
Pitch class usage profiles (PCP) from the literature.

In almost all cases reported here, keys are assumed to be
transpositionally equivalent, so the first (0th) entry is the tonic,
and no key-specific information is given.  The exception is QuinnWhite
which provides key-specific data.  In the key-specific case, we
instead store the distributions as a tuple of tuples of distributions
representing each individual key profile.

The profiles here provide the values exactly as reported in the
literature.  Where a profile does not sum to 1, an additional "_sum"
entry is provided with that normalisation.

The profiles appear below in approximately chronological order.
For reference, the alphabetical ordering is:

 -  AardenEssen
 -  AlbrechtShanahan
 -  BellmanBudge
 -  deClerqTemperley
 -  KrumhanslKessler
 -  KrumhanslSchmuckler
 -  PrinceSchumuckler
 -  QuinnWhite
 -  SappSimple
 -  TemperleyKostkaPayne
 -  TemperleyDeClerq
 -  Vuvan
 -  VuvanHughes

The variable **`source_list`** contains a list of all of these profiles.

Each profile is a dataclass object with the following attributes:

 - **`name`** (str) - the class name
 - **`about`** (str) - a short description
 - **`literature`** (str) - reference to the relevant publication
 - **`major`** (tuple[float]) - original weights for 12 pitch classes, major keys
   (in KrumhanslKessler, KrumhanslSchmuckler, AardenEssen, BellmanBudge,
    TemperleyKostkaPayne, Sapp, AlbrechtShanahan only)
 - **`major_sum`**  (tuple[float]) - weights that sum to 1, major keys
   (in KrumhanslKessler, KrumhanslSchmuckler, AardenEssen, BellmanBudge,
    TemperleyKostkaPayne, Sapp only)
 - **`minor`** (tuple[float]) - original weights for 12 pitch classes, minor keys
   (in KrumhanslKessler, KrumhanslSchmuckler, AardenEssen, BellmanBudge,
    TemperleyKostkaPayne, Sapp, AlbrechtShanahan only)
 - **`minor_sum`**  (tuple[float]) - weights that sum to 1, minor keys
   (in KrumhanslKessler, KrumhanslSchmuckler, AardenEssen, BellmanBudge,
    TemperleyKostkaPayne, Sapp only)
 - **`natural_minor`** (tuple[float]) - original weights, natural minor keys
   (in `Vuvan` only)
 - **`natural_minor_sum`** (tuple[float]) - weights that sum to 1, natural minor keys
   (in `Vuvan` only)
 - **`harmonic_minor`** (tuple[float]) - original weights, harmonic minor keys
   (in `Vuvan` only)
 - **`harmonic_minor_sum`** (tuple[float]) - weights that sum to 1, harmonic minor keys
   (in `Vuvan` only)
 - **`melodic_minor`** (tuple[float]) - original weights, melodic minor keys
   (in `Vuvan` only)
 - **`melodic_minor_sum`** (tuple[float]) - weights that sum to 1, melodic minor keys
   (in `Vuvan` only)
 - **`roots`** (tuple[float]) - original weights that sum to 1, chord roots in rock
   harmony (in DeClerqTemperley only)
 - **`melody_major`** (tuple[float]) - original weights that sum to 1, major melodies
   (in `TemperleyDeClerq` only)
 - **`melody_minor`** (tuple[float]) - original weights that sum to 1, minor melodies
   (in `TemperleyDeClerq` only)
 - **`harmony_major`** (tuple[float]) - original weights that sum to 1, major melodies
   (in `TemperleyDeClerq` only)
 - **`harmony_minor`** (tuple[float]) - original weights that sum to 1, minor melodies
   (in `TemperleyDeClerq` only)
 - **`downbeat_major`** (tuple[float]) - original weights,
   major on downbeats (in `PrinceSchumuckler` only)
 - **`downbeat_major_sum`** (tuple[float]) - original weights that sum to 1,
   major on downbeats (in `PrinceSchumuckler` only)
 - **`downbeat_minor`** (tuple[float]) - original weights,
   minor on downbeats (in `PrinceSchumuckler` only)
 - **`downbeat_minor_sum`** (tuple[float]) - original weights that sum to 1,
   minor on downbeats (in `PrinceSchumuckler` only)
 - **`all_beats_major`** (tuple[float]) - original weights,
   major on all beats (in `PrinceSchumuckler` only)
 - **`all_beats_major_sum`** (tuple[float]) - original weights that sum to 1,
   major on all beats (in `PrinceSchumuckler` only)
 - **`all_beats_minor`** (tuple[float]) - original weights,
   minor on all_beats (in `PrinceSchumuckler` only)
 - **`all_beats_minor_sum`** (tuple[float]) - original weights that sum to 1,
   minor on all_beats (in `PrinceSchumuckler` only)
 - **`major_all`** (tuple[float]) - original weights that sum to 1, all keys
   (in QuinnWhite only)
 - **`major_0`** (tuple[float]) - original weights that sum to 1, C Major
   (in QuinnWhite only)
 - **`major_1`** (tuple[float]) - original weights that sum to 1, C#/Db Major
   (in QuinnWhite only)
 - **`major_2`** (tuple[float]) - original weights that sum to 1, D Major
   (in QuinnWhite only)
 - **`major_3`** (tuple[float]) - original weights that sum to 1, D#/Eb Major
   (in QuinnWhite only)
 - **`major_4`** (tuple[float]) - original weights that sum to 1, E Major
   (in QuinnWhite only)
 - **`major_5`** (tuple[float]) - original weights that sum to 1, F Major
   (in QuinnWhite only)
 - **`major_6`** (tuple[float]) - original weights that sum to 1, F#/Gb Major
   (in QuinnWhite only)
 - **`major_7`** (tuple[float]) - original weights that sum to 1, G Major
   (in QuinnWhite only)
 - **`major_8`** (tuple[float]) - original weights that sum to 1, G#/Ab Major
   (in QuinnWhite only)
 - **`major_9`** (tuple[float]) - original weights that sum to 1, A Major,
   (in QuinnWhite only)
 - **`major_10`** (tuple[float]) - original weights that sum to 1, A#/Bb Major
   (in QuinnWhite only)
 - **`major_11`** (tuple[float]) - original weights that sum to 1, B Major
   (in QuinnWhite only)
 - **`minor_all`** (tuple[float]) - original weights that sum to 1, all keys
   (in QuinnWhite only)
 - **`minor_0`** (tuple[float]) - original weights that sum to 1, C Minor
   (in QuinnWhite only)
 - **`minor_1`** (tuple[float]) - original weights that sum to 1, C#/Db Minor
   (in QuinnWhite only)
 - **`minor_2`** (tuple[float]) - original weights that sum to 1, D Minor
   (in QuinnWhite only)
 - **`minor_3`** (tuple[float]) - original weights that sum to 1, D#/Eb Minor
   (in QuinnWhite only)
 - **`minor_4`** (tuple[float]) - original weights that sum to 1, E Minor
   (in QuinnWhite only)
 - **`minor_5`** (tuple[float]) - original weights that sum to 1, F Minor
   (in QuinnWhite only)
 - **`minor_6`** (tuple[float]) - original weights that sum to 1, F#/Gb Minor
   (in QuinnWhite only)
 - **`minor_7`** (tuple[float]) - original weights that sum to 1, G Minor
   (in QuinnWhite only)
 - **`minor_8`** (tuple[float]) - original weights that sum to 1, G#/Ab Minor
   (in QuinnWhite only)
 - **`minor_9`** (tuple[float]) - original weights that sum to 1, A Minor,
   (in QuinnWhite only)
 - **`minor_10`** (tuple[float]) - original weights that sum to 1, A#/Bb Minor
   (in QuinnWhite only)
 - **`minor_11`** (tuple[float]) - original weights that sum to 1, B Minor
   (in QuinnWhite only)
 - **`classical`** (tuple[float]) - original weights for 12 pitch classes,
   all classical keys
   (in VuvanHuges only)
 - **`classical_sum`**  (tuple[float]) - weights that sum to 1, all classical keys
   (in VuvanHuges only)
 - **`rock`** (tuple[float]) - original weights for 12 pitch classes, all rock keys
   (in VuvanHuges only)
 - **`rock_sum`**  (tuple[float]) - weights that sum to 1, all rock keys
   (in VuvanHuges only)

<small>**Author**: Mark Gotham, 2021, Huw Cheston, 2025</small>

REFERENCE
---------
Gotham et al. "What if the 'When' Implies the 'What'?". ISMIR, 2021
(see README.md)
"""

from dataclasses import dataclass
from typing import List, Tuple, Union

import numpy as np

import amads.algorithms.norm as norm
from amads.core.distribution import Distribution
from amads.core.pitch import CHROMATIC_NAMES


class PitchProfile(Distribution):
    """
     A set of weights for each pitch class denoting the expected frequency
     of pitches for collections of notes (typically songs, can be chords)
     of a given key.

     We provide methods to allow users to obtain or visualize this information
     in a useful state.

     Definitions:
     Define a canonical order of pitches as the order of pitches specified in
     relative chromatic degree.

     In our implementation, a pitch profile is a collection of pitch class
     distributions stored in a canonical form convenient for conversion
     into other useful forms, whether to provide methods in a useful state
     or for custom visualization.
     We store the pitch profile canonically in one of two forms:

     1. In the transpositionally equivalent case, we store the data as a
         list of 12 (float) weights beginning with the tonic.
     2. In the case of profiles that are not transpositionally equivalent,
         there is are weights for each key which are collectively stored
         as a list of 12 key profiles, each a list of 12 weights. Each
         profile begins with the tonic. Therefore `data[2][5]` represents
         the 5th weight (for pitch class G) in the 2nd key (D).

     For visualization, our design envisions the following use-cases:

     1. Compare and contrast data within a single PitchProfile object,
         especially between different key profiles beginning at their
         respecive tonic. Hence, our custom plot method allows a list
         of keys to plot the data side by side in a heatmap.
     2. Compare and contrast PitchProfile data with other pitch-class
         distributions, hence why we also satisfy the specifications
         for both singular plot and multiple plots from its parent
         class Distribution.

    Attributes
    ----------
    _profile_label : str, class attribute
        histogram label for 1-D histogram (x-axis), or representing the
        key of the current profile in the 2-D heatmap (y-axis)
    _data_cats_2d : List[int], class attribute
        data categories for the 2-D heatmap, which are labelled in terms
        of relative chromatic degree
    _data_labels : List[str], class attribute
        possible data labels, Relative Chromatic Degree for 2-D case (x-axis),
        and Weights for 1-D case (y-axis)
    """

    # possible pitches
    _profile_label = "Keys"
    _data_cats_2d = [f"{idx}" for idx in range(12)]
    _data_labels = ["Chromatic Scale Degree", "Weights"]

    def __init__(
        self,
        name: str,
        profile_tuple: Union[Tuple[float, ...], Tuple[Tuple[float, ...], ...]],
    ):
        if not PitchProfile._check_init_data_integrity(profile_tuple):
            raise ValueError(f"invalid profile tuple {profile_tuple}")
        profile_data = None
        profile_shape = None
        dist_type = None

        x_cats = None
        x_label = None
        y_cats = None
        y_label = None
        if isinstance(profile_tuple[0], float):
            profile_data = list(profile_tuple)
            profile_shape = [len(profile_data)]
            dist_type = "symmetric_key_profile"
            x_cats = CHROMATIC_NAMES
            x_label = PitchProfile._profile_label
            y_cats = None
            y_label = PitchProfile._data_labels[1]

        elif isinstance(profile_tuple[0], tuple):
            # convert data from tonic-first to non-rotated canonical order
            profile_data = [
                elem[-idx:] + elem[:-idx]  # type: ignore
                for idx, elem in enumerate(profile_tuple)
            ]
            profile_shape = [len(profile_data), len(profile_data[0])]
            dist_type = "asymmetric_key_profile"
            x_cats = PitchProfile._data_cats_2d
            x_label = PitchProfile._data_labels[0]
            y_cats = CHROMATIC_NAMES
            y_label = PitchProfile._profile_label

        else:
            raise ValueError(f"invalid profile tuple {profile_tuple}")
        super().__init__(
            name,
            profile_data,
            dist_type,
            profile_shape,
            x_cats,  # type: ignore (strings are ok)
            x_label,
            y_cats,  # type: ignore (strings are ok)
            y_label,
        )

    @classmethod
    def _check_init_data_integrity(
        cls, data: Union[Tuple[float, ...], Tuple[Tuple[float, ...], ...]]
    ) -> bool:
        """
        checks the integrity of the data tuple that is supplied in init
        """

        def allisfloat(data_tuple):
            return all(isinstance(elem, float) for elem in data_tuple)

        if len(data) != 12:
            return False
        is_valid_sym = allisfloat(data)
        if is_valid_sym:
            return True
        is_valid_asym = all(
            isinstance(elem, tuple) and len(elem) == 12 and allisfloat(elem)
            for elem in data
        )
        return is_valid_asym

    def normalize(self):
        """
        Normalize the pitch-class distributions within the PitchProfile.

        For each key, the sum of all weights in the corresponding key profile
        is normalized to 1.
        """
        if self.distribution_type == "symmetric_key_profile":
            self.data = norm.normalize(self.data, "sum").tolist()
            return self
        else:
            self.data = [
                norm.normalize(elem, "sum").tolist() for elem in self.data
            ]
            return self

    def key_to_weights(self, key: str) -> List[float]:
        """
        Given a key, computes the corresponding weights for the key profile.

        The key profile is rotated to the key as the tonic and organized in
        relative chromatic degree.

        Parameters
        ----------
        key: str
            pitch string denoting the key of the key profile data we want to retrieve

        Returns
        -------
        list[float]
            weights: list of 12 floats
        """
        key_idx = None
        try:  # C -> 0, C# -> 1, D -> 2, ..., B -> 11
            key_idx = CHROMATIC_NAMES.index(key.capitalize())
        except ValueError:
            raise ValueError(
                f"invalid key {key}, expected one of {CHROMATIC_NAMES}"
            )
        assert key_idx is not None
        assert key_idx >= 0 and key_idx < 12
        if self.distribution_type == "symmetric_key_profile":
            # symmetrical case
            return self.data
        elif self.distribution_type != "asymmetric_key_profile":
            ValueError(f"invalid distribution type {self.distribution_type}")
            # asymmetrical case
        return self.data[key_idx]

    def as_canonical_matrix(self) -> np.ndarray:
        """
        Computes a 12x12 matrix of the profile data.

            1. The i-th row corresponds to the key profile of the
               i-th chromatic degree

            2. Each row's weights begin from C and are ordered by
               relative chromatic degree)

        Returns
        -------
        np.ndarray
            a 12x12 numpy matrix of floats
        """
        assert self.dimensions[0] == 12
        if self.distribution_type == "symmetric_key_profile":
            # in this case, symmetric profile is transpositionally equivalent,
            # so for instance, in C# major, the pitch weights would be
            # transposed (rotated) as B -> C, C -> C#, C# -> D, ...
            profile_matrix = np.zeros((self.dimensions[0], self.dimensions[0]))
            for i in range(12):
                data = self.data[-i:] + self.data[:-i]
                profile_matrix[i] = data
            return profile_matrix
        else:
            assert self.distribution_type == "asymmetric_key_profile"
            assert self.dimensions == [12, 12]
            return np.array(self.data)

    # def profile_plot(
    #     self,
    #     color: str = Distribution.DEFAULT_BAR_COLOR,
    #     option: str = "bar",
    #     show: bool = True,
    #     fig: Optional[Figure] = None,
    #     ax: Optional[Axes] = None,
    # ) -> Figure:
    #     """
    #     Plot method overriding the original plot method in the parent class.
    #     Instead of using the behavior of the plot method in the parent class,
    #     this plot uses the default plot behavior (keys = None) from PitchProfile's
    #     plot_custom.

    #     See plot_custom for more details about argument behavior.

    #     Returns:
    #         Figure - A matplotlib figure object
    #     """
    #     # Figure/axes handling: either both `fig` and `ax` are provided, or
    #     # neither; in the latter case, create a new figure/axes pair.
    #     if fig is None:
    #         if ax is not None:
    #             raise ValueError("invalid figure/axis combination")
    #         fig, ax = plt.subplots()
    #     else:
    #         if ax is None:
    #             raise ValueError("invalid figure/axis combination")
    #     return self.plot_custom(
    #         keys=None, color=color, option=option, show=show, fig=fig, ax=ax
    #     )

    # def plot_custom(
    #     self,
    #     keys: Optional[List[str]] = None,
    #     color: str = Distribution.DEFAULT_BAR_COLOR,
    #     option: str = "bar",
    #     show: bool = True,
    #     fig: Optional[Figure] = None,
    #     ax: Optional[Axes] = None,
    # ) -> Figure:
    #     """
    #     custom plot method for PitchProfile.

    #     There are largely 3 cases of behavior, revolving around the keys variable.

    #     When keys is an empty list, return None.

    #     When the keys argument is a non-empty list of pitches:
    #     (1) Plot a heatmap where the i-th row is the key profile for the i-th key in keys.
    #     (2) Each row in the heatmap is ordered canonically (by relative chromatic degree)
    #     and begins with its corresponding tonic.

    #     The default custom plot option, or when keys argument is None, is as follows:
    #     (1) If the current PitchProfile is transpositionally equivalent, plot a bar graph
    #     or line graph depending on what is specified in the option argument with the
    #     specified color.
    #     (1) If the current PitchProfile is non-transpositionally equivalent, plot a grey
    #     2d heatmap

    #     Args:
    #     keys: Optional[list]
    #         a list of key pitches for which we want the corresponding pitch profiles
    #         to be visualized
    #     color: str
    #         (1) In the transpositionally equivalent case and keys is None,
    #         color is the color of the graph
    #         (2) Ignored otherwise
    #     option: str
    #         (1) In the transpositionally equivalent case and keys is None,
    #         "bar" for plotting a bar graph,
    #         or "line" or "plot" for plotting a line graph.
    #         (2) Ignored otherwise
    #     show: bool
    #         whether or not we want to display the plot before returning from this function
    #     fig, ax : matplotlib Figure and Axes
    #         Provide an existing figure/axes to draw on; if omitted, a new
    #         figure and axes are created.

    #     Returns:
    #         Figure - A matplotlib figure object.
    #     """
    #     # Figure/axes handling: either both `fig` and `ax` are provided, or
    #     # neither; in the latter case, create a new figure/axes pair.
    #     if fig is None:
    #         if ax is not None:
    #             raise ValueError("invalid figure/axis combination")
    #         fig, ax = plt.subplots()
    #     else:
    #         if ax is None:
    #             raise ValueError("invalid figure/axis combination")
    #     # similar logic to the plot method in distribution.py
    #     plot_keys = keys
    #     # in the default case, we have default presets to plot the data
    #     if plot_keys is None:
    #         # 1-D plot
    #         if self.distribution_type == "symmetric_key_profile":
    #             return super().plot(
    #                 fig=fig, ax=ax, color=color, option=option, show=show
    #             )
    #         else:
    #             plot_keys = CHROMATIC_NAMES
    #     assert isinstance(plot_keys, list)
    #     if not plot_keys:
    #         raise ValueError("empty keys list provided")

    #     plot_data = [self.key_to_weights(key) for key in plot_keys]
    #     # save original plot labels
    #     save_state = (self.x_categories, self.x_label, self.y_categories, self.y_label)

    #     self.x_categories = PitchProfile._data_cats_2d
    #     self.x_label = PitchProfile._data_labels[0]
    #     self.y_categories = plot_keys
    #     self.y_label = PitchProfile._profile_label
    #     fig = self._plot_2d(fig=fig, ax=ax, plot_data=plot_data)

    #     # restore original plot labels
    #     self.x_categories, self.x_label, self.y_categories, self.y_label = save_state

    #     if show:
    #         plt.show()
    #     return fig

    # def _plot_2d(
    #     self,
    #     fig: Figure,
    #     ax: Axes,
    #     plot_data: List[List[float]],
    # ) -> Figure:
    #     """Create a 2D heatmap of a non-transpositionally equivalent PitchProfile.
    #     This is currently only used as internal logic for plot_custom.
    #     Returns:
    #         Figure - A matplotlib figure object.
    #     """
    #     if plot_data is None or fig is None or ax is None:
    #         raise ValueError("invalid plotting arguments")
    #     cax = ax.imshow(
    #         plot_data, cmap="gray_r", interpolation="nearest", aspect="auto"
    #     )
    #     fig.colorbar(cax, ax=ax, label="Proportion")
    #     ax.set_xlabel(self.x_label)
    #     ax.set_ylabel(self.y_label)
    #     fig.suptitle(self.name)

    #     # Set x and y axis tick labels
    #     ax.set_xticks(range(len(self.x_categories)))
    #     ax.set_xticklabels(self.x_categories, rotation=45)
    #     ax.set_yticks(range(len(self.y_categories)))
    #     ax.set_yticklabels(self.y_categories)

    #     ax.invert_yaxis()
    #     return fig


@dataclass
class KeyProfile:
    """This is the base class for all key profiles.

    Attributes:
        name (str): the name of the profile
        literature (str): citations for the profile in the literature
        about (str): a longer description of the profile.
    """

    name: str = ""
    literature: str = ""
    about: str = ""

    def __getitem__(self, key: str):
        """This is added for (some) backwards compatibility, allowing objects
        to be accessed as dictionaries using bracket notation.

        Examples
        --------
            >>> kp = KrumhanslKessler()
            >>> kp["name"]
            'KrumhanslKessler'
        """
        try:
            return getattr(self, key)
        # Slightly nicer error handling
        except AttributeError:
            raise AttributeError(
                f"Key Profile '{self.__str__()}' has no attribute '{key}'"
            )

    def __str__(self) -> str:
        return self.name


@dataclass
class KrumhanslKessler(KeyProfile):
    name: str = "KrumhanslKessler"
    literature: str = (
        "Krumhansl and Kessler (1982). See also Krumhansl and Shepard (1979)"
    )
    about: str = (
        "Early PCP from psychological 'goodness of fit' tests using probe-tones"
    )
    major: PitchProfile = PitchProfile(
        "KrumhanslKessler.major",
        (
            6.35,
            2.23,
            3.48,
            2.33,
            4.38,
            4.09,
            2.52,
            5.19,
            2.39,
            3.66,
            2.29,
            2.88,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "KrumhanslKessler.minor",
        (
            6.33,
            2.68,
            3.52,
            5.38,
            2.6,
            3.53,
            2.54,
            4.75,
            3.98,
            2.69,
            3.34,
            3.17,
        ),
    )


@dataclass
class KrumhanslSchmuckler(KeyProfile):
    name: str = "KrumhanslSchmuckler"
    literature: str = "Krumhansl (1990)"
    about: str = (
        "Early case of key-estimation through matching usage with profiles"
    )
    major: PitchProfile = PitchProfile(
        "KrumhanslSchmuckler.major",
        (
            6.35,
            2.33,
            3.48,
            2.33,
            4.38,
            4.09,
            2.52,
            5.19,
            2.39,
            3.66,
            2.29,
            2.88,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "KrumhanslSchmuckler.minor",
        (
            6.33,
            2.68,
            3.52,
            5.38,
            2.6,
            3.53,
            2.54,
            4.75,
            3.98,
            2.69,
            3.34,
            3.17,
        ),
    )


@dataclass
class AardenEssen(KeyProfile):
    name: str = "AardenEssen"
    literature: str = "Aarden (2003) based on Schaffrath (1995)"
    about: str = "Folk melody transcriptions from the Essen collection"
    major: PitchProfile = PitchProfile(
        "AardenEssen.major",
        (
            17.7661,
            0.145624,
            14.9265,
            0.160186,
            19.8049,
            11.3587,
            0.291248,
            22.062,
            0.145624,
            8.15494,
            0.232998,
            4.95122,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "AardenEssen.minor",
        (
            18.2648,
            0.737619,
            14.0499,
            16.8599,
            0.702494,
            14.4362,
            0.702494,
            18.6161,
            4.56621,
            1.93186,
            7.37619,
            1.75623,
        ),
    )


@dataclass
class BellmanBudge(KeyProfile):
    name: str = "BellmanBudge"
    literature: str = (
        "Bellman (2005, sometimes given as 2006) after Budge (1943)"
    )
    about: str = "Chords in Western common practice tonality"
    major: PitchProfile = PitchProfile(
        "BellmanBudge.major",
        (
            16.8,
            0.86,
            12.95,
            1.41,
            13.49,
            11.93,
            1.25,
            20.28,
            1.8,
            8.04,
            0.62,
            10.57,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "BellmanBudge.minor",
        (
            18.16,
            0.69,
            12.99,
            13.34,
            1.07,
            11.15,
            1.38,
            21.07,
            7.49,
            1.53,
            0.92,
            10.21,
        ),
    )


@dataclass
class Temperley(KeyProfile):
    name: str = "Temperley"
    literature: str = (
        "Temperley (1999). What's Key for Key? The Krumhansl-Schmuckler Key-Finding Algorithm Reconsidered. Music Perception, 17, 65-100."
    )
    about: str = (
        "Psychological data revised - Temperley's revision of Krumhansl-Schmuckler profiles"
    )
    major: PitchProfile = PitchProfile(
        "Temperley.major",
        (
            5.0,
            2.0,
            3.5,
            2.0,
            4.5,
            4.0,
            2.0,
            4.5,
            2.0,
            3.5,
            1.5,
            4.0,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "Temperley.minor",
        (
            5.0,
            2.0,
            3.5,
            4.5,
            2.0,
            4.0,
            2.0,
            4.5,
            3.5,
            2.0,
            1.5,
            4.0,
        ),
    )


@dataclass
class TemperleyKostkaPayne(KeyProfile):
    name: str = "TemperleyKostkaPayne"
    literature: str = "Temperley (2007 and 2008)"
    about: str = (
        "Usage by section and excerpts from a textbook (Kostka & Payne)"
    )
    major: PitchProfile = PitchProfile(
        "TemperleyKostkaPayne.major",
        (
            0.748,
            0.06,
            0.488,
            0.082,
            0.67,
            0.46,
            0.096,
            0.715,
            0.104,
            0.366,
            0.057,
            0.4,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "TemperleyKostkaPayne.minor",
        (
            0.712,
            0.084,
            0.474,
            0.618,
            0.049,
            0.46,
            0.105,
            0.747,
            0.404,
            0.067,
            0.133,
            0.33,
        ),
    )


@dataclass
class Sapp(KeyProfile):
    name: str = "Sapp"
    literature: str = "Sapp (PhD thesis, 2011)"
    about: str = (
        "Simple set of scale degree intended for use with Krumhansl Schmuckler (above)"
    )
    major: PitchProfile = PitchProfile(
        "Sapp.major",
        (2.0, 0.0, 1.0, 0.0, 1.0, 1.0, 0.0, 2.0, 0.0, 1.0, 0.0, 1.0),
    )
    minor: PitchProfile = PitchProfile(
        "Sapp.minor",
        (2.0, 0.0, 1.0, 1.0, 0.0, 1.0, 0.0, 2.0, 1.0, 0.0, 1.0, 0.0),
    )


@dataclass
class Vuvan(KeyProfile):
    name: str = "Vuvan"
    literature: str = "Vuvan et al. (2011)"
    about: str = "Different profiles for natural, harmonic, and melodic minors"
    natural_minor: PitchProfile = PitchProfile(
        "Vuvan.natural_minor",
        (
            5.08,
            3.03,
            3.73,
            4.23,
            3.64,
            3.85,
            3.13,
            5.29,
            4.43,
            3.95,
            5.26,
            3.99,
        ),
    )
    harmonic_minor: PitchProfile = PitchProfile(
        "Vuvan.harmonic_minor",
        (
            4.62,
            2.63,
            3.74,
            4.23,
            3.63,
            3.81,
            4.15,
            5.21,
            4.77,
            3.95,
            3.79,
            5.3,
        ),
    )
    melodic_minor: PitchProfile = PitchProfile(
        "Vuvan.melodic_minor",
        (
            4.75,
            3.26,
            3.76,
            4.46,
            3.49,
            4.09,
            3.67,
            5.08,
            4.14,
            4.43,
            4.51,
            4.91,
        ),
    )


@dataclass
class DeClerqTemperley(KeyProfile):
    name: str = "DeClerqTemperley"
    literature: str = "deClerq and Temperley (Popular Music, 2011)"
    about: str = "Chord roots (specifically) in rock harmony."
    roots: PitchProfile = PitchProfile(
        "DeClerqTemperley.roots",
        (
            0.328,
            0.005,
            0.036,
            0.026,
            0.019,
            0.226,
            0.003,
            0.163,
            0.04,
            0.072,
            0.081,
            0.004,
        ),
    )


@dataclass
class TemperleyDeClerq(KeyProfile):
    name: str = "TemperleyDeClerq"
    literature: str = "Temperley and deClerq (JNMR, 2013)"
    about: str = """Rock music and a distinction between melody and harmony.
               distributions as reported in Vuvan and Hughes (2021, see below)
               following personal correspondence with Temperley."""
    major: PitchProfile = PitchProfile(
        "TemperleyDeClerq.major",
        (
            0.223,
            0.001,
            0.158,
            0.015,
            0.194,
            0.071,
            0.002,
            0.169,
            0.003,
            0.119,
            0.008,
            0.035,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "TemperleyDeClerq.minor",
        (
            0.317,
            0.001,
            0.09,
            0.159,
            0.046,
            0.097,
            0.007,
            0.131,
            0.009,
            0.047,
            0.087,
            0.009,
        ),
    )
    harmony_major: PitchProfile = PitchProfile(
        "TemperleyDeClerq.harmony_major",
        (
            0.231,
            0.002,
            0.091,
            0.004,
            0.149,
            0.111,
            0.004,
            0.193,
            0.004,
            0.126,
            0.011,
            0.076,
        ),
    )
    harmony_minor: PitchProfile = PitchProfile(
        "TemperleyDeClerq.harmony_minor",
        (
            0.202,
            0.006,
            0.102,
            0.127,
            0.047,
            0.113,
            0.005,
            0.177,
            0.046,
            0.051,
            0.09,
            0.034,
        ),
    )


@dataclass
class AlbrechtShanahan(KeyProfile):
    name: str = "AlbrechtShanahan"
    literature: str = "Albrecht and Shanahan (Music Perception, 2013)"
    about: str = """Partial pieces for more stable within-key environment.
               Note that the two pairs of distributions reported in the appendix are identical"""
    major: PitchProfile = PitchProfile(
        "AlbrechtShanahan.major",
        (
            0.238,
            0.006,
            0.111,
            0.006,
            0.137,
            0.094,
            0.016,
            0.214,
            0.009,
            0.08,
            0.008,
            0.081,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "AlbrechtShanahan.minor",
        (
            0.220,
            0.006,
            0.104,
            0.123,
            0.019,
            0.103,
            0.012,
            0.214,
            0.062,
            0.022,
            0.061,
            0.052,
        ),
    )


@dataclass
class PrinceSchumuckler(KeyProfile):
    name: str = "PrinceSchumuckler"
    literature: str = "Prince and Schmuckler (Music Perception, 2014)"
    about: str = """Distinction between downbeat and all beats.
               Note they also provide profiles for metrical position usage."""
    downbeat_major: PitchProfile = PitchProfile(
        "PrinceSchumuckler.downbeat_major",
        (
            1.0,
            0.088610811,
            0.569205361,
            0.140888014,
            0.615384615,
            0.481864956,
            0.140888014,
            0.976815092,
            0.156831608,
            0.433398971,
            0.122721209,
            0.427237502,
        ),
    )
    downbeat_minor: PitchProfile = PitchProfile(
        "PrinceSchumuckler.downbeat_minor",
        (
            1.0,
            0.127885863,
            0.516472114,
            0.640207523,
            0.174189364,
            0.537483787,
            0.160311284,
            0.989883268,
            0.426588846,
            0.172114137,
            0.430350195,
            0.286381323,
        ),
    )
    major: PitchProfile = PitchProfile(
        "PrinceSchumuckler.major",
        (
            0.919356471,
            0.114927991,
            0.729198287,
            0.144709771,
            0.697021822,
            0.525970522,
            0.214762724,
            1.0,
            0.156143546,
            0.542952545,
            0.142399406,
            0.541215555,
        ),
    )
    minor: PitchProfile = PitchProfile(
        "PrinceSchumuckler.minor",
        (
            0.874192439,
            0.150655606,
            0.637256776,
            0.697274361,
            0.162238618,
            0.62471807,
            0.167131771,
            1.0,
            0.47788524,
            0.212622807,
            0.467754884,
            0.298711724,
        ),
    )


@dataclass
class QuinnWhite(KeyProfile):
    name: str = "QuinnWhite"
    literature: str = "Quinn and White (Music Perception 2017)"
    about: str = "Separate profiles for each key"
    # this used to be major_all, but is the symmetrical version of
    # the major key profile in QuinnWhite
    major: PitchProfile = PitchProfile(
        "QuinnWhite.major",
        (
            0.172,
            0.014,
            0.107,
            0.011,
            0.160,
            0.099,
            0.018,
            0.231,
            0.017,
            0.059,
            0.016,
            0.093,
        ),
    )
    # instead of ordering them by circle of fifths, we order the distributions by
    # incrementing key number the non-transpositionally equivalent distributions
    # represent instead
    major_asym: PitchProfile = PitchProfile(
        "QuinnWhite.major_asym",
        (
            (
                0.174,
                0.014,
                0.112,
                0.010,
                0.160,
                0.100,
                0.016,
                0.230,
                0.015,
                0.058,
                0.016,
                0.096,
            ),
            (
                0.173,
                0.016,
                0.103,
                0.014,
                0.165,
                0.091,
                0.020,
                0.234,
                0.018,
                0.055,
                0.018,
                0.093,
            ),
            (
                0.175,
                0.013,
                0.113,
                0.010,
                0.155,
                0.102,
                0.017,
                0.227,
                0.017,
                0.061,
                0.015,
                0.094,
            ),
            (
                0.173,
                0.015,
                0.108,
                0.011,
                0.160,
                0.098,
                0.019,
                0.231,
                0.016,
                0.059,
                0.015,
                0.094,
            ),
            (
                0.171,
                0.013,
                0.108,
                0.010,
                0.161,
                0.106,
                0.015,
                0.229,
                0.019,
                0.059,
                0.016,
                0.092,
            ),
            (
                0.173,
                0.014,
                0.108,
                0.009,
                0.161,
                0.096,
                0.018,
                0.231,
                0.016,
                0.063,
                0.016,
                0.095,
            ),
            (
                0.166,
                0.018,
                0.096,
                0.014,
                0.170,
                0.085,
                0.019,
                0.240,
                0.019,
                0.062,
                0.020,
                0.091,
            ),
            (
                0.175,
                0.013,
                0.108,
                0.011,
                0.156,
                0.102,
                0.017,
                0.233,
                0.016,
                0.058,
                0.019,
                0.091,
            ),
            (
                0.171,
                0.014,
                0.099,
                0.013,
                0.164,
                0.099,
                0.018,
                0.235,
                0.016,
                0.064,
                0.015,
                0.093,
            ),
            (
                0.174,
                0.014,
                0.108,
                0.012,
                0.160,
                0.101,
                0.017,
                0.232,
                0.018,
                0.059,
                0.014,
                0.093,
            ),
            (
                0.169,
                0.016,
                0.107,
                0.013,
                0.158,
                0.099,
                0.020,
                0.230,
                0.017,
                0.060,
                0.017,
                0.096,
            ),
            (
                0.167,
                0.014,
                0.106,
                0.014,
                0.164,
                0.100,
                0.018,
                0.233,
                0.021,
                0.054,
                0.020,
                0.089,
            ),
        ),
    )
    minor: PitchProfile = PitchProfile(
        "QuinnWhite.minor",
        (
            0.170,
            0.012,
            0.115,
            0.149,
            0.013,
            0.095,
            0.027,
            0.211,
            0.074,
            0.024,
            0.026,
            0.085,
        ),
    )
    minor_asym: PitchProfile = PitchProfile(
        "QuinnWhite.minor_asym",
        (
            (
                0.170,
                0.012,
                0.118,
                0.141,
                0.011,
                0.098,
                0.026,
                0.212,
                0.074,
                0.024,
                0.023,
                0.091,
            ),
            (
                0.168,
                0.014,
                0.112,
                0.152,
                0.014,
                0.093,
                0.028,
                0.212,
                0.078,
                0.022,
                0.026,
                0.082,
            ),
            (
                0.172,
                0.010,
                0.118,
                0.158,
                0.012,
                0.092,
                0.023,
                0.211,
                0.067,
                0.027,
                0.027,
                0.084,
            ),
            (
                0.168,
                0.014,
                0.111,
                0.152,
                0.015,
                0.087,
                0.032,
                0.213,
                0.073,
                0.025,
                0.027,
                0.082,
            ),
            (
                0.174,
                0.012,
                0.114,
                0.149,
                0.013,
                0.098,
                0.029,
                0.202,
                0.076,
                0.025,
                0.023,
                0.087,
            ),
            (
                0.167,
                0.011,
                0.117,
                0.141,
                0.012,
                0.093,
                0.026,
                0.217,
                0.077,
                0.024,
                0.025,
                0.089,
            ),
            (
                0.172,
                0.013,
                0.109,
                0.149,
                0.014,
                0.091,
                0.029,
                0.215,
                0.075,
                0.025,
                0.028,
                0.081,
            ),
            (
                0.174,
                0.011,
                0.116,
                0.152,
                0.014,
                0.094,
                0.028,
                0.208,
                0.069,
                0.027,
                0.026,
                0.081,
            ),
            (
                0.168,
                0.014,
                0.106,
                0.151,
                0.014,
                0.093,
                0.028,
                0.212,
                0.076,
                0.022,
                0.030,
                0.085,
            ),
            (
                0.175,
                0.010,
                0.114,
                0.149,
                0.012,
                0.096,
                0.025,
                0.217,
                0.073,
                0.021,
                0.026,
                0.083,
            ),
            (
                0.164,
                0.011,
                0.113,
                0.150,
                0.014,
                0.095,
                0.033,
                0.205,
                0.078,
                0.027,
                0.027,
                0.083,
            ),
            (
                0.164,
                0.012,
                0.120,
                0.144,
                0.013,
                0.102,
                0.024,
                0.208,
                0.074,
                0.022,
                0.028,
                0.088,
            ),
        ),
    )


@dataclass
class VuvanHughes(KeyProfile):
    name: str = "VuvanHughes"
    literature: str = "Vuvan and Hughes (Music Perception 2021)"
    about: str = "A comparison of Classical and Rock music."
    classical: PitchProfile = PitchProfile(
        "VuvanHughes.classical",
        (
            5.38,
            2.65,
            3.39,
            3.01,
            3.62,
            3.96,
            2.83,
            4.93,
            2.9,
            3.38,
            2.91,
            3.03,
        ),
    )
    rock: PitchProfile = PitchProfile(
        "VuvanHughes.rock",
        (
            5.34,
            3.33,
            3.73,
            3.39,
            3.95,
            3.99,
            2.82,
            4.54,
            2.9,
            3.21,
            2.88,
            2.71,
        ),
    )


source_list = (
    AardenEssen,
    AlbrechtShanahan,
    BellmanBudge,
    DeClerqTemperley,
    KrumhanslKessler,
    KrumhanslSchmuckler,
    PrinceSchumuckler,
    QuinnWhite,
    Sapp,
    Temperley,
    TemperleyKostkaPayne,
    TemperleyDeClerq,
    Vuvan,
    VuvanHughes,
)


aarden_essen = AardenEssen()
albrecht_shanahan = AlbrechtShanahan()
bellman_budge = BellmanBudge()
declerq_temperley = DeClerqTemperley()
krumhansl_kessler = KrumhanslKessler()
krumhansl_schmuckler = KrumhanslSchmuckler()
prince_schmuckler = PrinceSchumuckler()
quinn_white = QuinnWhite()
sapp = Sapp()
temperley = Temperley()
temperley_kostka_payne = TemperleyKostkaPayne()
temperley_de_clerq = TemperleyDeClerq()
vuvan = Vuvan()
vuvan_hughes = VuvanHughes()



from amads.core.basics import Note


def test_ties():
    note_1 = Note(onset=0.0, duration=1.0, pitch=60)
    note_2 = Note(onset=1.0, duration=1.0, pitch=60)

    note_1.tie = note_2

    # I'm not sure if this is the desired behavior,
    # but it's what the code looks like it's trying to do
    assert note_1.duration == 1.0
    assert note_1.tied_duration == 2.0



"""
Shared functionality for basic operations on vectors
(applicable to more than one musical parameter).
Includes
transformations (e.g., `rotate`) and
checks (e.g., `is_rotation_equivalent`) on vectors.

<small>**Author**: Mark Gotham</small>
"""

__author__ = "Mark Gotham"

from itertools import combinations
from typing import List, Optional, Sequence, Union

from amads.core.vectors_sets import (
    is_indicator_vector,
    multiset_to_vector,
    pairwise_differences,
)

# ----------------------------------------------------------------------------

# Transformations


def rotate(
    vector: Union[tuple[int, ...], list[int]], steps: Union[int, None] = None
) -> Union[tuple[int, ...], list[int]]:
    """
    this serves equivalently for
    "phase shifting" of rhythm and
    "transposition" of pitch.

    Parameters
    ----------
    vector : Union[tuple[int, ...], list[int]]
        Any tuple or list of any elements.
        We expect to work with a list of integers representing a vector.
    steps: Optional[int]
        How many steps to rotate.
        Or, equivalently, the nth index of the input list becomes the 0th index of the new.
        If unspecified, use the half cycle: int(cycle_length / 2).

    Returns
    -------
    Union[tuple[int, ...], list[int]]
        The input (tuple or list), rotated. Same length.

    Examples
    --------

    >>> start = (0, 1, 2, 3)
    >>> rotate(start, 1)
    (1, 2, 3, 0)

    >>> rotate(start, -1)
    (3, 0, 1, 2)

    >>> rotate(start) # note no steps specified
    (2, 3, 0, 1)

    """
    if steps is None:
        steps = int(len(vector) / 2)

    return vector[steps:] + vector[:steps]


def mirror(
    vector: Union[list, tuple], index_of_symmetry: Union[int, None] = None
) -> Union[list, tuple]:
    """
    Reverse a vector (or any ordered iterable).

    Parameters
    ----------
    vector: tuple
        The tuple accepts any ordered succession of any elements.
        We expect integers representing a vector, but do not enforce it.
    index_of_symmetry: Union[int, None]
        Defaults to None, in which case, standard reflection of the
        form `[::-1]`. Alternatively, specify an index to rotate about,
        e.g., for the reverse function in convolution use 0. This is
        equivalent to mirror and rotation. See notes at [`rotate`]
        [amads.core.vector_transforms_checks.rotate].

    Returns
    -------
    Union[tuple[int, ...], list[int]]
        The input (tuple or list), mirrored. Same length.

    Examples
    --------
    >>> test_case = (0, 1, 2, 3, 4, 5)
    >>> mirror(test_case)
    (5, 4, 3, 2, 1, 0)

    >>> mirror(test_case, index_of_symmetry=0)
    (0, 5, 4, 3, 2, 1)

    >>> mirror(test_case, index_of_symmetry=1)
    (1, 0, 5, 4, 3, 2)

    We will often use this for a 12-element indicator vector.
    >>> c_vector = (1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0)
    >>> mirror(c_vector, index_of_symmetry=0)
    (1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0)

    """
    if index_of_symmetry is not None:
        rotated = (
            vector[index_of_symmetry::-1] + vector[-1:index_of_symmetry:-1]
        )
    else:
        rotated = vector[::-1]
    return rotated


def complement(indicator_vector: tuple[int, ...]) -> tuple[int, ...]:
    """
    Provide the complement of an indicator vector.

    Returns
    -------
    tuple[int]

    Examples
    --------
    >>> complement((1, 0, 1, 0))
    (0, 1, 0, 1)
    """
    if not is_indicator_vector(indicator_vector):
        raise ValueError(
            "This is to be called only on binary tuples representing indicator vectors."
        )
    return tuple(1 - x for x in indicator_vector)


def change_cycle_length(
    start_vector: tuple,
    destination_length: int,
    return_indices_not_indicator: bool = True,
) -> tuple:
    """
    Change the cycle length of a vector by mapping each point to the nearest equivalent in the new length.

    Examples
    --------
    >>> tresillo = (1, 0, 0, 1, 0, 0, 1, 0)
    >>> change_cycle_length(tresillo, 9)
    (0, 3, 7)

    >>> change_cycle_length(tresillo, 12)
    (0, 4, 9)

    >>> change_cycle_length(start_vector=tresillo, destination_length=9,
    ...                     return_indices_not_indicator=False)
    (1, 0, 0, 1, 0, 0, 0, 1, 0)

    """
    start_indices = indicator_to_indices(start_vector)
    start_len = len(start_vector)
    new_indices = [
        round(destination_length * i / start_len) for i in start_indices
    ]
    if return_indices_not_indicator:
        return tuple(new_indices)
    else:
        return indices_to_indicator(
            new_indices, indicator_length=destination_length
        )


# ----------------------------------------------------------------------------

# Checks


def is_rotation_equivalent(vector_a: tuple, vector_b: tuple) -> bool:
    """
    Given two vectors, test for rotation equivalence.
    This is applicable to indicator vectors, interval sequences, and more
    as long as the user compares like with like.

    Examples
    --------

    Indicator:

    >>> is_rotation_equivalent((1, 0, 0), (0, 1, 0))
    True

    >>> is_rotation_equivalent((1, 0, 0), (1, 1, 0))
    False

    Intervals:

    >>> is_rotation_equivalent((3, 3, 2), (3, 2, 3))
    True

    >>> is_rotation_equivalent((3, 3, 2), (2, 3, 2))
    False

    Trivial case:
    >>> is_rotation_equivalent((1, 0, 0), (1, 0, 0))
    True

    """
    if vector_a == vector_b:
        return True

    vector_length = len(vector_a)
    if len(vector_b) != vector_length:
        raise ValueError(
            f"The vectors must be of the same length (currently {vector_length} and {len(vector_b)}."
        )

    for steps in range(1, vector_length):
        if vector_a == rotate(vector_b, steps):
            return True

    return False


def is_maximally_even(indicator_vector: tuple) -> bool:
    """
    Checks if an indicator vector (tuple of 0s and 1s) is maximally even,
    meaning the 1s and 0s are as evenly distributed as possible.

    Specifically, this works simply by converting to intervals and running the following basic checks:
    First, there must be no more than 2 interval types.
    If there is only 1 type, True (perfectly even).
    If there are 2 types, then those two must differ in value by 1.

    Parameters
    ----------
    indicator_vector: tuple[int, ...]
        An indicator_vector (only 0s and/or 1s).

    Returns
    -------
    bool
        True if the pattern is maximally even, False otherwise.

    Examples
    --------
    >>> is_maximally_even((1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0))
    True
    >>> is_maximally_even((1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0))
    False
    >>> is_maximally_even((0, 0, 1, 0, 0, 1, 0, 0, 1))
    True
    >>> is_maximally_even((1, 1, 0, 0, 1))
    False
    >>> is_maximally_even((0, 1, 0, 1, 0, 1))
    True

    This works with the `indicator_to_interval` function.
    Let's look at those representations for the bembé cycle and a comparison case.

    >>> bembé = (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1)
    >>> indicator_to_interval(bembé, adjacent_not_all=True)
    (2, 2, 1, 2, 2, 2, 1)

    >>> indicator_to_interval(bembé, adjacent_not_all=True, sequence_not_vector=False)
    (2, 5, 0, 0, 0, 0)

    >>> indicator_to_interval(bembé, adjacent_not_all=False)
    (2, 5, 4, 3, 6, 1)

    >>> is_maximally_even(bembé)
    True

    For our comparison case, we create another cycle that also has:
    a) 7 elements in a 12-unit cycle,

    >>> not_bembé =  (1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1)
    >>> not_bembé.count(1)
    7

    >>> len(not_bembé)
    12

    b) the same first-order interval set of 5 x 2s and 2 x 1s,

    >>> indicator_to_interval(not_bembé, adjacent_not_all=True, sequence_not_vector=False)
    (2, 5, 0, 0, 0, 0)

    ... but for which the those 2 x 1s are together

    >>> indicator_to_interval(not_bembé, adjacent_not_all=True)
    (2, 2, 2, 2, 2, 1, 1)

    ... making it not maximally even.

    >>> is_maximally_even(not_bembé)
    False

    """
    if not is_indicator_vector(indicator_vector):
        raise ValueError(
            f"The `indicator_vector` argument (set as {indicator_vector}) should be an indicator vector."
        )

    k = indicator_vector.count(1)
    n = len(indicator_vector)

    prototype_k_in_n = indices_to_indicator(tuple(max_even_k_in_n(k, n)), n)

    return is_rotation_equivalent(prototype_k_in_n, indicator_vector)


# ----------------------------------------------------------------------------

# Other


def max_even_k_in_n(k: int, n: int) -> set[int]:
    """
    Make a maximally even pattern of k elements in an n-length cycle.
    Normally, we would expect k to be less than n, but this is not strictly required.
    Larger values of k simply produce duplicate entries in n that are ignored in the
    returned set.

    Examples
    --------
    >>> max_even_k_in_n(3, 8)
    {0, 3, 5}

    >>> max_even_k_in_n(7, 12)
    {0, 2, 3, 5, 7, 9, 10}

    >>> max_even_k_in_n(5, 16)
    {0, 3, 6, 10, 13}

    """
    return set([round(n * i / k) for i in range(k)])


def rotation_distinct_patterns(
    vector_patterns: tuple[tuple, ...]
) -> tuple[tuple, ...]:
    """
    Given two or more vectors of the same length,
    test rotation equivalence among them.
    Return the list of rotation distinct pattens:
    the returned patterns are not rotation equivalent to each other,
    but all tested rhythms (in the argument) are rotation equivalent
    to one of those returned patterns.

    Examples
    --------
    Here are ten canonical 12-unit bell pattern rhythms:

    >>> Soli = (2, 2, 2, 2, 1, 2, 1)
    >>> Tambú = (2, 2, 2, 1, 2, 2, 1)
    >>> Bembé = (2, 2, 1, 2, 2, 2, 1)
    >>> Bembé_2 = (1, 2, 2, 1, 2, 2, 2)
    >>> Yoruba = (2, 2, 1, 2, 2, 1, 2)
    >>> Tonada = (2, 1, 2, 1, 2, 2, 2)
    >>> Asaadua = (2, 2, 2, 1, 2, 1, 2)
    >>> Sorsonet = (1, 1, 2, 2, 2, 2, 2)
    >>> Bemba = (2, 1, 2, 2, 2, 1, 2)
    >>> Ashanti = (2, 1, 2, 2, 1, 2, 2)
    >>> ten_tuples = (Asaadua, Ashanti, Bemba, Bembé, Bembé_2, Soli, Sorsonet, Tambú, Tonada, Yoruba)

    Collectively, they have 3 distinct patterns.

    >>> len(rotation_distinct_patterns(ten_tuples))
    3

    """
    return_values = [vector_patterns[0]]  # At least one
    for index in range(1, len(vector_patterns)):
        for prototype in return_values:
            if is_rotation_equivalent(prototype, vector_patterns[index]):
                break
        else:
            return_values.append(vector_patterns[index])

    return tuple(return_values)


def indicator_to_indices(
    vector: Union[list[int], tuple[int, ...]], wrap: bool = False
) -> tuple:
    """
    Simple mapping from an indicator vector for where events fall,
    to a tuple of the corresponding indices.

    Parameters
    ----------
    vector: Union[list[int], tuple[int, ...]]
        an indicator vector

    wrap: bool
        if true, the first element of `vector` is appended to `vector`
        before computing indices, so if the element is non-zero, the
        length of `vector` will be appear in the result.

    Returns
    -------
    tuple[int, ...]
        a tuple containing the indices for which `vector` is non-zero.

    Examples
    --------
    >>> bembé = (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1) # adjacent only
    >>> indicator_to_indices(bembé)
    (0, 2, 4, 5, 7, 9, 11)

    >>> indicator_to_indices(bembé, wrap=True)
    (0, 2, 4, 5, 7, 9, 11, 12)

    >>> shiko = (1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0) # adjacent only
    >>> indicator_to_indices(shiko)
    (0, 4, 6, 10, 12)

    >>> indicator_to_indices(shiko, wrap=True)
    (0, 4, 6, 10, 12, 16)

    """
    if wrap:
        vector = vector + vector[:1]

    return tuple(i for i, v in enumerate(vector) if v)


# TODO in vectors_sets?
def indices_to_indicator(
    indices_vector: Union[list[int], tuple[int, ...]],
    indicator_length: Optional[int] = None,
) -> tuple:
    """
    Simple mapping from indices to indicator vector.

    Parameters
    ----------
    indices_vector: Union[list[int], tuple[int, ...]]
        A vector of indices (0, 2, 4, 5, 7, 9, 11).
        Monotonic increase is expected but not required.
    indicator_length: Optional[int]
        optionally specify the length of the output indicator vector.
        If not specified, we use the highest index in the indices vector.

    Examples
    --------
    Round trip
    >>> bembé = (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1)
    >>> indices = indicator_to_indices(bembé)
    >>> indices
    (0, 2, 4, 5, 7, 9, 11)

    No `indicator_length` is needed for default

    >>> round_trip = indices_to_indicator(indices)
    >>> round_trip
    (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1)

    >>> round_trip == bembé
    True

    The `indicator_length` can, however, be specified:
    >>> indices_to_indicator(indices, indicator_length=12)
    (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1)

    And the `indicator_length` can extend the indicator as needed:
    >>> indices_to_indicator(indices, indicator_length=14)
    (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1, 0, 0)

    """
    if indicator_length is None:
        indicator_length = max(indices_vector) + 1
    index_set = set(indices_vector)
    return tuple(1 if i in index_set else 0 for i in range(indicator_length))


def indicator_to_interval(
    vector: Union[list[int], tuple[int, ...]],
    wrap: bool = True,
    adjacent_not_all: bool = True,
    sequence_not_vector: bool = True,
) -> tuple:
    """
    Given a vector (assumed to be indicator)
    convert from 1/0 at each index to the intervals between the 1s.

    Parameters
    ----------
    vector: Union[list[int], tuple[int, ...]]
        an indicator vector representing positions.
    wrap: bool
        wrap the cycle, duplicating the first element at the end to include
        that (possible) interval.
    adjacent_not_all: bool
        If True, form the set of intervals between pairs of adjacent positions.
        If False, form the set of intervals between all pairs.
    sequence_not_vector: bool
        In the case of adjacent intervals, True means express the result as an
        interval sequence, and False means compute an interval vector containing
        the counts of intervals of size 1, 2, 3, etc. If `adjacent_not_all` is
        False (meaning all pairs), the result is always an interval vector.

    Examples
    --------
    Example 1:
    >>> bembé = (1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 0, 1)

    Adjacent intervals expressed as a sequence:
    >>> indicator_to_interval(bembé)
    (2, 2, 1, 2, 2, 2, 1)

    Wrap is optional:
    >>> indicator_to_interval(bembé, wrap=False)
    (2, 2, 1, 2, 2, 2)

    Adjacent intervals expressed as an interval vector:
    >>> indicator_to_interval(bembé, sequence_not_vector=False)
    (2, 5, 0, 0, 0, 0)

    All distances (not just the adjacent pairs), which is necessarily
    expressed as an interval vector:
    >>> indicator_to_interval(bembé, adjacent_not_all=False)
    (2, 5, 4, 3, 6, 1)

    Example 2:
    >>> shiko = (1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0)

    Adjacent intervals expressed as a sequence:
    >>> indicator_to_interval(shiko)
    (4, 2, 4, 2, 4)

    Adjacent intervals expressed as an interval vector:
    >>> indicator_to_interval(shiko, sequence_not_vector=False)
    (0, 2, 0, 3, 0, 0, 0, 0)

    All distances (not just the adjacent pairs), which is necessarily
    expressed as an interval vector:
    >>> indicator_to_interval(shiko, adjacent_not_all=False)
    (0, 2, 0, 3, 0, 4, 0, 1)

    """
    if adjacent_not_all and wrap:
        vector = vector + vector[:1]

    indices = indicator_to_indices(vector)

    if adjacent_not_all:
        sequence = tuple(
            [indices[i + 1] - indices[i] for i in range(len(indices) - 1)]
        )
        if sequence_not_vector:
            return sequence
        else:
            ics = sequence
    else:  # all
        ics = [p[1] - p[0] for p in combinations(indices, 2)]

    vector_length = len(vector)
    half_vector_length = int(vector_length / 2)
    interval_vector = [0] * half_vector_length

    for ic in ics:
        if ic < 0:
            ic *= -1
        if ic > half_vector_length:
            ic = vector_length - ic
        interval_vector[ic - 1] += 1
    return tuple(interval_vector)


def interval_sequence_to_indices(
    interval_sequence_vector: Union[list[int], tuple[int, ...]],
    wrap: bool = False,
) -> tuple[int, ...]:
    """
    Given an interval sequence vector, convert to indices.

    Parameters
    ----------
    interval_sequence_vector: Union[list[int], tuple[int, ...]]
        a vector of distances (intervals) between adjacent positions
        that are used.
    wrap: bool
        If True, include the index after the end of the sequence

    Returns
    -------
    tuple[int, ...]
        A vector containing the positions (indices) that are used.

    Examples
    --------
    >>> interval_sequence_to_indices((3, 3, 2), wrap=False)  # Default
    (0, 3, 6)

    >>> interval_sequence_to_indices((3, 3, 2), wrap=True)
    (0, 3, 6, 8)

    """
    indices = [0]
    count = 0
    if not wrap:
        interval_sequence_vector = interval_sequence_vector[:-1]
    for i in interval_sequence_vector:
        count += i
        indices.append(count)
    return tuple(indices)


def is_monotonic(numbers: Sequence) -> bool:
    """
    Assert that a list of numbers is monotonically increasing.
    Returns True if so, raises ValueError at the first failure point.

    Please note that edge case behaviour (e.g., 0, None, raises) may change.

    Parameters
    ----------
        numbers: A sequence (list, tuple, ...) of numeric values (integers, floats, ...).

    Examples
    --------
    >>> is_monotonic([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    True
    >>> is_monotonic([1, 2, 3, 4, 5, 7, 6, 8, 9, 10])
    Traceback (most recent call last):
    ValueError: Data must be monotonically increasing: value 7 at index 5 is not less than 6.

    """
    for i in range(len(numbers) - 1):
        if numbers[i] >= numbers[i + 1]:
            raise ValueError(
                "Data must be monotonically increasing: value "
                f"{numbers[i]} at index {i} is not less than {numbers[i + 1]}."
            )
    return True


def indices_to_interval_sequence(
    indices: Union[list[int], tuple[int, ...]],
    wrap: bool = True,
    mod: int = 12,
) -> tuple[int, ...]:
    """
    Convert a sequence of indices (any increasing integers, n >= 2)
    into a sequence of the differences, mod 12 (including the difference between the last and the first items)

    Parameters
    ----------
    indices: Union[list[int], tuple[int, ...]]
        A vector containing the positions (indices) that are used.
    wrap: bool
        If True, include the index after the end of the sequence
    mod: int
        The modulo value for use when wrapping around

    Returns
    -------
    tuple[int, ...]
        A vector of distances (intervals) between adjacent positions

    Examples
    --------

    >>> indices_to_interval_sequence([1, 2, 3])
    (1, 1, 10)

    >>> indices_to_interval_sequence([1, 2, 3], wrap=False)
    (1, 1)

    >>> indices_to_interval_sequence([1, 2, 6])
    (1, 4, 7)

    >>> start = (3, 3, 2)
    >>> corresponding_indices = interval_sequence_to_indices(start, wrap=False)  # Default
    >>> corresponding_indices
    (0, 3, 6)

    >>> end = indices_to_interval_sequence(corresponding_indices, mod=8)
    >>> end
    (3, 3, 2)

    >>> start == end
    True

    """
    is_monotonic(indices)
    if len(indices) < 2:
        raise ValueError(
            f"Starting indices have 2 or more items, currently {indices} is of length {len(indices)}."
        )

    diffs = []
    for i in range(len(indices) - 1):
        diffs.append(indices[i + 1] - indices[i])

    if wrap:
        diffs += [(indices[0] - indices[-1]) % mod]

    return tuple(diffs)


def interval_sequence_to_indicator(
    interval_sequence_vector: Union[list[int], tuple[int, ...]]
) -> tuple[int, ...]:
    """
    Given an interval sequence vector, convert to indicator.

    Parameters
    ----------
    interval_sequence_vector: Union[list[int], tuple[int, ...]]
        the sequence of distances (intervals) between adjacent positions
        that are used.

    Returns
    -------
    tuple[int, ...]
        An indicator vector representing all positions, where used
        positions contain the value 1.

    Examples
    --------
    >>> interval_sequence_to_indicator((3, 3, 2))
    (1, 0, 0, 1, 0, 0, 1, 0)

    """
    indices = interval_sequence_to_indices(interval_sequence_vector, wrap=False)
    indicator = indices_to_indicator(
        indices, indicator_length=sum(interval_sequence_vector)
    )
    return tuple(indicator)


def saturated_subsequence_repetition(
    sequence: Union[List[int], tuple[int, ...]],
    all_rotations: bool = True,
    subsequence_period: Optional[int] = None,
) -> List[List[int]]:
    """
    Check if a sequence contains a repeated subsequence such that
    the subsequence saturates the whole (no sequence items "left over").
    This is broadly equivalent to a "periodic sequence", with the additional
    constraint of saturatation.

    This function is a wrapper for an abstraction provided at
    [saturated_subsequence_repetition]
    [amads.core.vector_transforms_checks.saturated_subsequence_repetition].

    Parameters
    ----------
    sequence: Union[list[int], tuple[int, ...]]
        A vector for event positions in the cycle time span.
    all_rotations: bool
        If True, check all rotations of the sequence.
    subsequence_period: Optional[int]
        If specified, check only that period length; otherwise,
        check all factors of n.

    Returns
    -------
    Union[List[List[int]], None]
        A list of all subsequences that, if repeated, can form the input
        `sequence`. If `all_rotations`, then also include subsequences
        that can repeat to form any rotations of `sequence`. If
        `sequence` is not periodic, return None.

    Examples
    --------
    >>> test_sequence = [1, 2, 1, 2, 1, 2, 1, 2]

    All rotations, all subsequence lengths:

    >>> saturated_subsequence_repetition(test_sequence, all_rotations=True)
    [[1, 2], [2, 1], [1, 2, 1, 2], [2, 1, 2, 1]]

    No rotations, all subsequence lengths:

    >>> saturated_subsequence_repetition(test_sequence, all_rotations=False)
    [[1, 2], [1, 2, 1, 2]]

    All rotations, subsequence length fixed at 2:

    >>> saturated_subsequence_repetition(test_sequence, subsequence_period=2, all_rotations=True)
    [[1, 2], [2, 1]]

    All rotations, subsequence length fixed at 4:

    >>> saturated_subsequence_repetition(test_sequence, subsequence_period=4, all_rotations=True)
    [[1, 2, 1, 2], [2, 1, 2, 1]]

    No rotations, subsequence length fixed at 2:
    >>> saturated_subsequence_repetition(test_sequence, subsequence_period=2, all_rotations=False)
    [[1, 2]]

    No rotations, subsequence length fixed at 4:
    >>> saturated_subsequence_repetition(test_sequence, subsequence_period=4, all_rotations=False)
    [[1, 2, 1, 2]]

    """
    if subsequence_period is not None:
        subsequence_periods = [subsequence_period]
    else:
        subsequence_periods = [
            length
            for length in range(1, len(sequence) // 2 + 1)
            if len(sequence) % length == 0
        ]

    subsequences = []

    for period in subsequence_periods:
        subsequence = sequence[:period]
        if subsequence not in subsequences:
            if all(
                sequence[i : i + period] == subsequence
                for i in range(0, len(sequence), period)
            ):
                subsequences.append(subsequence)

        if all_rotations:
            for j in range(
                1, period
            ):  # sic, from 1 (0 is done) and only up to the length of the subsequence
                this_sequence = rotate(sequence, j)
                subsequence = this_sequence[:period]
                if subsequence not in subsequences:
                    if all(
                        this_sequence[k : k + period] == subsequence
                        for k in range(0, len(this_sequence), period)
                    ):
                        subsequences.append(subsequence)

    return subsequences


def convolve(
    vector_1: Sequence, vector_2: Sequence, return_indicator: bool = False
) -> tuple:
    """
    Convolution combines two vectors,
    with a scalar product of corresponding entries
    where they are rotated relative to one another by k steps.

    Parameters
    ----------
    vector_1: Sequence
        A vector of numeric values.
    vector_2: Sequence
        Another vector of numeric values, must be the same length as `vector_1` (otherwise, raises).
    return_indicator: bool
        If True, return an indicator vector (only 1s and 0s).

    Returns
    -------
    tuple
        The convolution of the two input vectors.
        If `return_indicator` is True, values are reduced to 1s and 0s.

    Examples
    --------
    This function is agnostic wrt musical parameter;
    the following example as applied to pitch is illustrative.

    >>> c_major_triad = (1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0)
    >>> min_7_dyad = (1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0)
    >>> convolve(c_major_triad, min_7_dyad)
    (1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0)

    """
    number_elements = len(vector_1)
    if len(vector_2) != number_elements:
        raise ValueError(
            "The lengths of two vectors must match. "
            f"Currently {vector_1} ({number_elements}) "
            f"and {vector_2} ({len(vector_2)})."
        )

    combined_list = []
    for k in range(number_elements):
        combined_list.append(
            sum(
                vector_1[i] * vector_2[(k - i) % number_elements]
                for i in range(number_elements)
            )
        )

    if return_indicator:
        combined_list = [1 if x > 0 else 0 for x in combined_list]
    return tuple(combined_list)


def interval_function(
    vector_1: Sequence[int],
    vector_2: Sequence[int],
    return_vector: bool = True,
) -> tuple[int, ...]:
    """
    All directed intervals between two vectors.
    The interval _function_ was introduced to music theory by Lewin, 2001 [1]

    [1] David Lewin, Special Cases of the Interval Function between Pitch-Class Sets X and Y,
    Journal of Music Theory (2001) 45 (1): 1–29. https://doi.org/10.2307/3090647

    Parameters
    ----------
    vector_1: Sequence[int]
        A sequence representing a starting condition.
    vector_2: Sequence[int]
        A sequence representing a following condition.
    return_vector: bool
        If True (default), return a count vector indexed by interval size
        (length ``max_interval + 1``).
        If False, return the raw multiset of directed intervals mod 12.

    Returns
    -------
    tuple[int, ...]
        If ``return_vector`` is True, a 13-element tuple where index ``i``
        is the count of directed interval ``i`` (mod 12) from ``vector_1``
        to ``vector_2``.
        If ``return_vector`` is False, the raw multiset of directed intervals.

    Examples
    --------
    >>> start_set = [1, 2, 6]
    >>> end_set = [1, 5, 7]

    With `return_vector=False` we get the pairwise differences as a set multiset of intervals in the range 0–12.
    >>> interval_function(start_set, end_set, return_vector=False)
    (0, 4, 6, 11, 3, 5, 7, 11, 1)

    With `return_vector=True` we get an interval vector with usage per position (/interval).
    For example, note that we have two instance of interval 11 in the above.
    >>> interval_function(start_set, end_set, return_vector=True)
    (1, 1, 0, 1, 1, 1, 1, 1, 0, 0, 0, 2)

    """
    differences = pairwise_differences(vector_1, vector_2, modulo=12)

    if return_vector:
        differences = multiset_to_vector(differences, max_index=11)

    return tuple(differences)


# ------------------------------------------------------------------------

if __name__ == "__main__":
    import doctest

    doctest.testmod()



import math
from typing import Iterator, Optional, Union

from amads.core.basics import Pitch
from amads.io.readscore import read_score, valid_score_extensions


def dir_to_collection(filenames: list[str]):
    """
    Converts a list of music filenames to a dictionary where keys
    are filenames and values are corresponding Score objects.

    Parameters
    ----------
    filenames : list(str)
        List of filenames to process.

    Returns
    -------
    dict
        A dictionary mapping filenames to Score objects.
    """
    scores = {}

    for file in filenames:
        if any(file.endswith(ext) for ext in valid_score_extensions):
            try:
                score = read_score(file)
                scores[file] = score
            except Exception as e:
                print(f"Error processing file {file}: {e}")
        else:
            print(f"Unsupported file format: {file}")

    return scores


def _hz_to_key_num_single(hz: float, do_round: bool = True) -> float:
    """Helper function for hz_to_key_num"""
    key_num = 69 + 12 * math.log2(hz / 440.0)
    return round(key_num) if do_round else key_num


def hz_to_key_num(
    hertz: Union[float, list[float]], do_round: bool = True
) -> Union[float, list[float]]:
    """
    Converts a frequency in Hertz to the corresponding MIDI key number.

    Parameters
    ----------
    hertz : Union(float, list(float))
        The frequency or list of frequencies in Hertz.

    do_round : bool
        Perform rounding to the nearest integer key_num.

    Returns
    -------
    Union(float, list(float))
        The corresponding MIDI key number(s).

    Examples
    --------
    >>> hz_to_key_num(440.0)
    69
    >>> hz_to_key_num(260.0, False)
    59.89209719404554
    >>> hz_to_key_num([440, 260], True)
    [69, 60]
    """

    if isinstance(hertz, list):
        return [_hz_to_key_num_single(hz, do_round) for hz in hertz]
    else:
        return _hz_to_key_num_single(hertz, do_round)


def hz_to_pitch(
    hertz: Union[float, list[float]], round: bool = True
) -> Union[Pitch, list[Pitch]]:
    """
    Converts a frequency to a Pitch object.

    Parameters
    ----------
    hertz : Union(float, list(float))
        The frequency or list of frequencies in Hertz.

    round : bool
        Perform rounding to the nearest integer key_num.

    Returns
    -------
    Union(Pitch, list(Pitch))
        The corresponding Pitch objects.

    Examples
    --------
    >>> hz_to_pitch(440)
    Pitch(name='A4', key_num=69)
    """
    key_nums = hz_to_key_num(hertz, round)
    if isinstance(key_nums, list):
        return [Pitch(kn) for kn in key_nums]
    else:
        return Pitch(key_nums)


def key_num_to_hz(
    key_num: Union[float, Pitch, list[Union[float, Pitch]]]
) -> Union[float, list[float]]:
    """
    Converts a Pitch object or MIDI key number to the corresponding
    frequency in Hertz.

    Parameters
    ----------
    key_num : Union(Pitch, float, list(Union(Pitch, float)))
        The Pitch object(s) or MIDI key number(s).

    Returns
    -------
    Union(float, list(float))
        The corresponding frequency in Hertz.

    Examples
    --------
    >>> key_num_to_hz(69)
    440.0
    >>> key_num_to_hz([Pitch("A5"), 60])
    [880.0, 261.6255653005986]
    """

    def key_num_to_hz_single(k):
        if isinstance(k, Pitch):
            key_num = k.key_num
        else:
            key_num = k
        return 440.0 * 2 ** ((key_num - 69) / 12)

    if isinstance(key_num, list):
        return [key_num_to_hz_single(k) for k in key_num]
    else:
        return key_num_to_hz_single(key_num)


def key_num_to_name(n, detail="nameoctave"):
    """
    Converts key numbers to key names (text).

    Parameters
    ----------
    n : Union(int, list(int))
        The key numbers.
    detail : Optional(str)
        `'nameonly'` for just the note name (e.g., `'C#'`),
        `'nameoctave'` for note name with octave (e.g., `'C#4'`) (default).

    Returns
    -------
    Union(str, list(str))
        The corresponding key names.

    Examples
    --------
    >>> key_num_to_name(61)
    'C#4'

    >>> key_num_to_name(61, detail="nameonly")
    'C#'
    """

    def key_num_to_name_single(k):
        pitch = Pitch(k)
        if detail == "nameonly":
            # Handles sharps, flats, and naturals correctly
            return pitch.name
        elif detail == "nameoctave":
            return pitch.name_with_octave  # Includes note name and octave
        else:
            raise ValueError(
                "Invalid detail option. Use 'nameonly' or " "'nameoctave'."
            )

    if isinstance(n, list):
        return [key_num_to_name_single(k) for k in n]
    else:
        return key_num_to_name_single(n)


def sign(x: float) -> int:
    """
    Get the sign of a numeric value as -1, 0, or +1.

    Returns
    ----------
    int
        -1 if `x` < 0, 0 if `x` == 0, 1 if `x` > 0

    Examples
    --------
    >>> sign(-15)
    -1

    >>> sign(-1)
    -1

    >>> sign(-0.5)
    -1

    >>> sign(-0)
    0

    >>> sign(+0)
    0

    >>> sign(+0.5)
    1

    >>> sign(15.2)
    1

    >>> sign(None) is None
    True
    """
    if x is None:
        return None  # type: ignore
    else:
        return bool(x > 0) - bool(x < 0)


def float_range(
    start: float, end: Optional[float], step: float
) -> Iterator[float]:
    """Generate a range of floats.

    Similar to Python's built-in range() function but supports floating
    point numbers. If end is None, generates an infinite sequence.

    Parameters
    ----------
    start : float
        The starting value of the range
    end : float or None
        The end value of the range (exclusive). If None, generates an
        infinite sequence
    step : float
        The increment between values

    Yields
    ------
    float
        The next value in the sequence
    """
    curr = start
    while end is None or curr < end:
        yield curr
        curr += step


def check_python_package_installed(package_name: str):
    """
    Check if a Python package is installed, raise error if not.

    Parameters
    ----------
    package_name : str
        Name of the package to check

    Raises
    ------
    ImportError
        If package is not installed, with message suggesting pip install
    """
    try:
        __import__(package_name)
    except ImportError:
        raise ImportError(
            f"Package '{package_name}' is required but not installed. "
            f"Please install it using: pip install {package_name}"
        )



from collections import Counter
from typing import Dict

import numpy as np

from amads.algorithms.mtype_tokenizer import FantasticTokenizer
from amads.algorithms.ngrams import NGramCounter
from amads.core.basics import Score
from amads.melody.contour.huron_contour import HuronContour
from amads.melody.contour.interpolation_contour import InterpolationContour
from amads.melody.contour.parsons_contour import ParsonsContour
from amads.melody.contour.polynomial_contour import PolynomialContour
from amads.melody.contour.step_contour import StepContour
from amads.melody.segment import fantastic_segmenter

__author__ = "David Whyatt"


def fantastic_pitch_features(score: Score) -> Dict:
    """Extract pitch features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract pitch features from.

    Returns
    -------
    Dict
        A dictionary of pitch features.
        Dictionary keys:
            - pitch_range: The range of pitches in the melody.
            - pitch_std: The standard deviation of the pitches in the melody.
            - pitch_entropy: A variant of the Shannon entropy of the pitches in the melody.
    """
    notes = score.get_sorted_notes()

    pitches = [note.pitch.key_num for note in notes]

    pitch_range = max(pitches) - min(pitches)
    pitch_std = np.std(pitches)

    # Calculate pitch entropy using Shannon's formula
    # First get frequency distribution of pitches
    pitch_counts = Counter(pitches)
    total_pitches = len(pitches)

    # Calculate relative frequencies
    pitch_freqs = {
        p: count / total_pitches for p, count in pitch_counts.items()
    }

    # Calculate entropy using the formula from the FANTASTIC toolbox
    pitch_entropy = -sum(
        f * np.log2(f) for f in pitch_freqs.values()
    ) / np.log2(24)

    return {
        "pitch_range": pitch_range,
        "pitch_std": pitch_std,
        "pitch_entropy": pitch_entropy,
    }


def fantastic_pitch_interval_features(score: Score) -> Dict:
    """Extract pitch interval features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract pitch interval features from.

    Returns
    -------
    Dict
        A dictionary of pitch interval features.
        Dictionary keys:
            - absolute_interval_range: The range of absolute pitch intervals in the melody.
            - mean_absolute_interval: The mean of the absolute pitch intervals in the melody.
            - std_absolute_interval: The standard deviation of the absolute pitch intervals in the melody.
            - modal_interval: The modal absolute pitch interval in the melody.
            - interval_entropy: A variant of the Shannon entropy of the absolute pitch intervals in the melody.
    """
    notes = score.get_sorted_notes()

    pitches = [note.pitch.key_num for note in notes]
    # Fantastic defines intervals by looking forwards
    intervals = [pitches[i + 1] - pitches[i] for i in range(len(pitches) - 1)]
    # and then always uses the absolute value
    abs_intervals = [abs(interval) for interval in intervals]

    absolute_interval_range = max(abs_intervals) - min(abs_intervals)
    mean_absolute_interval = np.mean(abs_intervals)
    std_absolute_interval = np.std(abs_intervals)
    modal_interval = max(set(abs_intervals), key=abs_intervals.count)

    # Calculate interval entropy using Shannon's formula
    # First get frequency distribution of intervals
    interval_counts = Counter(abs_intervals)
    total_intervals = len(abs_intervals)

    # Calculate relative frequencies
    interval_freqs = {
        i: count / total_intervals for i, count in interval_counts.items()
    }

    # Calculate entropy using the formula from the FANTASTIC toolbox
    # Note that the maximum number of different intervals is instead 23 here
    interval_entropy = -sum(
        f * np.log2(f) for f in interval_freqs.values()
    ) / np.log2(23)

    return {
        "absolute_interval_range": absolute_interval_range,
        "mean_absolute_interval": mean_absolute_interval,
        "std_absolute_interval": std_absolute_interval,
        "modal_interval": modal_interval,
        "interval_entropy": interval_entropy,
    }


def fantastic_duration_features(score: Score) -> Dict:
    """Extract duration features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract duration features from.

    Returns
    -------
    Dict
        A dictionary of duration features.
        Dictionary keys:
    """

    raise NotImplementedError("Not implemented yet")
    # Fantastic mixes between duration in seconds and metrical tatums
    # Until issue #75 is resolved, we will not implement this function for now


def fantastic_global_features(score: Score) -> Dict:
    """Extract global extension features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract global extension features from.

    Returns
    -------
    Dict
        A dictionary of global extension features.
        Dictionary keys:
    """

    raise NotImplementedError("Not implemented yet")
    # As before, until issue #75 is resolved, we will not implement this function


def fantastic_step_contour_features(score: Score) -> Dict:
    """Extract step contour features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract step contour features from.

    Returns
    -------
    Dict
        A dictionary of step contour features.
        Dictionary keys:
            - global_variation: The global variation of the step contour.
            - global_direction: The global direction of the step contour.
            - local_variation: The local variation of the step contour.
    """
    notes = score.get_sorted_notes()

    # Extract pitches and times for contour calculation
    pitches = [note.pitch.key_num for note in notes]
    durations = [note.duration for note in notes]

    sc = StepContour(pitches, durations)

    return {
        "global_variation": sc.global_variation,
        "global_direction": sc.global_direction,
        "local_variation": sc.local_variation,
    }


def fantastic_interpolation_contour_features(score: Score) -> Dict:
    """Extract interpolation contour features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract interpolation contour features from.

    Returns
    -------
    Dict
        A dictionary of interpolation contour features.
        Dictionary keys:
            - global_direction: The global direction of the interpolation contour.
            - mean_gradient: The mean gradient of the interpolation contour.
            - gradient_std: The standard deviation of the gradient of the interpolation contour.
            - direction_changes: The number of direction changes in the interpolation contour.
            - class_label: The class label of the interpolation contour.
    """
    notes = score.get_sorted_notes()

    # Extract pitches and times for contour calculation
    pitches = [note.pitch.key_num for note in notes]
    times = [note.onset for note in notes]

    # Calculate contour
    ic = InterpolationContour(pitches, times, method="fantastic")

    return {
        # Interpolation contour features
        "global_direction": ic.global_direction,
        "mean_gradient": ic.mean_gradient,
        "gradient_std": ic.gradient_std,
        "direction_changes": ic.direction_changes,
        "class_label": ic.class_label,
    }


def fantastic_parsons_contour_features(
    score: Score, character_dict: Dict = None, initial_asterisk: bool = False
) -> Dict:
    """Extract Parsons contour features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract Parsons contour features from.

    Returns
    -------
    Dict
        A dictionary of Parsons contour features.
        Dictionary keys:
            - interval_sequence: The interval sequence of the melody.
            - interval_sequence_sign: A representation of the direction of the interval sequence
                using -1, 0, and 1 to represent down, repeat, and up intervals respectively.
            - as_string: The Parsons contour as a string, using the characters u, r, and d
                to represent up, repeat, and down intervals respectively.
    """

    notes = score.get_sorted_notes()

    pitches = [note.pitch.key_num for note in notes]
    pc = ParsonsContour(
        pitches,
        character_dict=character_dict,
        initial_asterisk=initial_asterisk,
    )

    return {
        "interval_sequence": pc.interval_sequence,
        "interval_sequence_sign": pc.interval_sequence_sign,
        "as_string": pc.as_string,
    }


def fantastic_polynomial_contour_features(score: Score) -> Dict:
    """Extract polynomial contour features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract polynomial contour features from.

    Returns
    -------
    Dict
        A dictionary of polynomial contour coefficients.
        Dictionary keys:
            - coefficients: The coefficients of the polynomial contour.
    """

    pc = PolynomialContour(score)

    return {
        "coefficients": pc.coefficients,
    }


def fantastic_huron_contour_features(score: Score) -> Dict:
    """Extract Huron contour features from a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to extract Huron contour features from.

    Returns
    -------
    Dict
        A dictionary of Huron contour features.
        Dictionary keys:
            - first_pitch: The first pitch of the melody.
            - mean_pitch: The mean pitch of the melody.
            - last_pitch: The last pitch of the melody.
            - first_to_mean: The difference between the first and mean pitch.
            - mean_to_last: The difference between the mean and last pitch.
            - contour_class: The class of the Huron contour.
    """
    notes = score.get_sorted_notes()

    pitches = [note.pitch.key_num for note in notes]
    times = [note.onset for note in notes]

    hc = HuronContour(pitches, times)

    return {
        "first_pitch": hc.first_pitch,
        "mean_pitch": hc.mean_pitch,
        "last_pitch": hc.last_pitch,
        "first_to_mean": hc.first_to_mean,
        "mean_to_last": hc.mean_to_last,
        "contour_class": hc.contour_class,
    }


def fantastic_count_mtypes(
    score: Score, segment: bool, phrase_gap: float, units: str
) -> NGramCounter:
    """Count M-Types in a melody.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to count M-Types in.
    segment : bool
        Whether to segment the melody into phrases.
    phrase_gap : float
        The minimum IOI gap to consider a new phrase.
    units : str
        The units of the phrase gap, either "seconds" or "quarters".

    Returns
    -------
    NGramCounter
        An NGramCounter object containing the counts of M-Types.
        This allows for the computation of the complexity measures, either
        by accessing the properties of the NGramCounter object or by using
        the `fantastic_mtype_summary_features` function.
    """
    if segment:
        segments = fantastic_segmenter(score, phrase_gap, units)
    else:
        segments = [score]

    counter = NGramCounter()
    tokenizer = FantasticTokenizer()

    all_tokens = []
    for phrase in segments:
        tokens = tokenizer.tokenize(phrase)
        all_tokens.extend(tokens)

    counter.count_ngrams(all_tokens, n=[1, 2, 3, 4, 5])

    return counter


def fantastic_mtype_summary_features(
    score: Score, segment: bool, phrase_gap: float, units: str
) -> Dict:
    """Count M-Types in a melody and compute summary features.

    This function provides an easy way to get all the complexity measures
    at once.

    <small>**Author**: David Whyatt</small>

    Parameters
    ----------
    score : Score
        The score to count M-Types in.
    segment : bool
        Whether to segment the melody into phrases.
    phrase_gap : float
        The minimum IOI gap to consider a new phrase.
    units : str
        The units of the phrase gap, either "seconds" or "quarters".

    Returns
    -------
    Dict
        A dictionary of summary features.
        Dictionary keys:

            - mean_entropy: The mean entropy of the M-Types.
            - mean_productivity: The mean productivity of the M-Types.
            - yules_k: The mean Yules K statistic.
            - simpsons_d: The mean Simpson's D statistic.
            - sichels_s: The mean Sichels S statistic.
            - honores_h: The mean Honores H statistic.
    """
    mtype_counts = fantastic_count_mtypes(score, segment, phrase_gap, units)

    return {
        "mean_entropy": mtype_counts.mean_entropy,
        "mean_productivity": mtype_counts.mean_productivity,
        "yules_k": mtype_counts.yules_k,
        "simpsons_d": mtype_counts.simpsons_d,
        "sichels_s": mtype_counts.sichels_s,
        "honores_h": mtype_counts.honores_h,
    }



"""
This module provides the `pitch_class_distribution_2` function.

Implements the Midi Toolbox `pcdist2` function.

Original doc: https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=81.
"""

from typing import cast

from amads.core.basics import Note, Part, Score
from amads.core.distribution import Distribution
from amads.core.histogram import Histogram2D
from amads.core.pitch import CHROMATIC_NAMES
from amads.pitch.pcdist1 import duraccent


def update_pcd(pcd: list[list[float]], notes: list[Note], weighted: bool):
    """Updates the pitch-class distribution matrix based on the given notes.

    Serves as a helper function for `pitch_class_distribution_2`

    Args:
        pcd (list[list[float]]): The pitch-class distribution matrix to be
                                 updated.
        notes (list[Note]): The list of notes to process.
        weighted (bool, optional): If True, the pitch-class distribution is
                                   weighted by note durations.
    """
    prev = None
    for note in notes:
        if prev:
            pc_curr = note.pitch_class
            pc_prev = prev.pitch_class

            if weighted:
                pcd[pc_prev][pc_curr] += prev.duration * note.duration
            else:
                pcd[pc_prev][pc_curr] += 1

        prev = note


def pitch_class_distribution_2(
    score: Score,
    name: str = "Pitch Class Distribution",
    weighted: bool = True,
    miditoolbox_compatible: bool = False,
) -> Distribution:
    """Returns the 2nd order pitch-class distribution of a musical score.

    Parameters
    ----------
    score : Score
        The musical score to analyze
    name : str
        Name for the distribution; plot title if plotted.
    weighted : bool, optional
        If True, weight the pitch-class distribution by note durations
        in seconds that are modified according to Parncutt's durational
        accent model (1994), by default True.
    miditoolbox_compatible : bool
        Matlab MIDI Toolbox avoids zero division by dividing counts
        by the total count plus (1e-12 times the number of bins).
        True enables this behavior. Default is False, which simply skips
        division when the total count is zero (this also returns a
        zero matrix when the count is zero).

    Returns
    -------
    Distribution
        A 12x12 distribution representing the transition probabilities of
        each pitch class (C, C#, D, D#, E, F, F#, G, G#, A, A#, B). If the
        score is empty, the function returns a distribution with all
        elements set to zero.
    """
    score = cast(Score, score.merge_tied_notes())
    if weighted:
        score.convert_to_seconds()  # need seconds for duraccent calculation
    bin_centers = [float(i) for i in range(12)]  # 25 bins from -12 to +12
    x_categories = CHROMATIC_NAMES
    h = Histogram2D(bin_centers, None, "linear", False)

    # do not count transitions from one part to the next
    for p in score.find_all(Part):
        part: Part = cast(Part, p)
        prev_bin = None
        prev_pc = None
        prev_dur = None
        for n in part.find_all(Note):
            note: Note = cast(Note, n)
            pc = note.pitch_class
            if weighted and prev_pc is not None:
                dur = duraccent(note)
                w = prev_dur + dur  # type: ignore
                prev_dur = dur
            else:
                w = 1.0
            prev_bin = h.add_point_2d(prev_pc, pc, w, prev_bin)
            prev_pc = pc

    if miditoolbox_compatible:  # miditoolbox "normalization"
        total = sum(h.bins) + len(h.bins) * 1e-12
        h.bins = [b / total for b in h.bins]
    else:  # normalize normally
        h.normalize()

    return Distribution(
        name,
        h.bins,
        "pitch_class_transition",
        [12, 12],
        x_categories,  # type: ignore
        "Current Pitch Class",
        x_categories,  # type: ignore
        "Previous Pitch Class",
    )



from typing import List

from amads.core.basics import Part, Score


def fantastic_segmenter(
    score: Score, phrase_gap: float, units: str
) -> List[Score]:
    """Segment melody into phrases based on IOI gaps.
    Parameters
    ----------
    score : Score
        Score object containing melody to segment
    phrase_gap : float
        The minimum IOI gap to consider a new phrase
    units : str
        The units of the phrase gap, either "seconds" or "quarters"

    Returns
    -------
    list[Score]
        List of Score objects representing phrases
    """
    assert units in ["seconds", "quarters"]
    if units == "seconds":
        raise NotImplementedError(
            "Seconds are not yet implemented, see issue #75: "
            "https://github.com/music-computing/amads/issues/75"
        )
    if units == "quarters":
        # Extract notes from score
        notes = score.get_sorted_notes()

        # Create a dictionary to store IOI information
        ioi_data = {}

        # Calculate IOIs
        for i, note in enumerate(notes):
            # Initialize entry for this note
            ioi_data[note] = None

            # first note has no IOI by convention
            if i > 0:
                ioi_data[note] = note.onset - notes[i - 1].onset
            else:
                ioi_data[note] = None

        phrases = []
        current_phrase = []
        for note in notes:
            # Check whether we need to make a new phrase
            need_new_phrase = (
                len(current_phrase) > 0
                and ioi_data[note]
                is not None  # Check current note's IOI instead of previous note
                and ioi_data[note] > phrase_gap
            )
            if need_new_phrase:
                # Create new score for the phrase
                phrase_score = Score(onset=0, duration=None)
                part = Part(
                    parent=None, onset=0, duration=None
                )  # parent=None is required
                start_time = current_phrase[0].onset
                # Adjust note timings relative to phrase start
                for phrase_note in current_phrase:
                    # make a parentless copy of the note so we can adjust its onset
                    # before inserting it into the new part in proper time order
                    new_note = phrase_note.insert_copy_into(None)
                    new_note.onset -= start_time
                    part.insert(new_note)
                phrase_score.insert(part)  # This will set the parent
                phrases.append(phrase_score)
                current_phrase = []
            current_phrase.append(note)

        # Append final phrase
        if len(current_phrase) > 0:
            phrase_score = Score(onset=0, duration=None)
            part = Part(
                parent=None, onset=0, duration=None
            )  # parent=None is required
            start_time = current_phrase[0].onset
            for phrase_note in current_phrase:
                new_note = phrase_note.insert_copy_into(None)
                new_note.onset -= start_time
                part.insert(new_note)
            phrase_score.insert(part)  # This will set the parent
            phrases.append(phrase_score)

        return phrases



"""
Assuming key of C, find a score's mode based on key profiles using key_cc.

This function is primarily used to estimate the mode, an attribute of a given
KeyProfile collection. The key of C is assumed, and cross-correlation with
profiles for other keys are ignored.

Reference
---------
https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=65
"""

from typing import List, Optional

import amads.pitch.key.profiles as prof
from amads.core.basics import Score
from amads.pitch.key.key_cc import key_cc


def keymode(
    score: Score,
    profile: prof.KeyProfile = prof.krumhansl_kessler,
    attribute_names: Optional[List[str]] = ["major", "minor"],
    salience_flag: bool = False,
) -> List[str]:
    """
    Find the mode based on cross-correlation values.

    Returns the list of mode(s) whose profile(s) have a maximal
    cross-correlation with the score's pitch distribution.

    Parameters
    ----------
    score: Score
        The musical score to analyze.
    profile: Profile
        collection of profile data for different modes (attributes)
    attribute_names: Optional[List[str]]
        List of attribute names that denote the particular PitchProfiles
        within the KeyProfile and generally indicate different modes.
        See profiles.py for more details.
    salience_flag: bool
        Indicate whether we want to turn on salience weights in key_cc
        which is used to compute the cross-correlations.

    Returns
    -------
    List[str]
        List of attribute names that have maximal cross-correlation with
        the score's profile (usually, length will be 1).

    See Also
    --------
    key_cc
    """

    # This algorithm is not very efficient: It computes 12 correlations
    # for each mode, but only uses one. Then, it iterates through the
    # results, once to find the maximum, and again to form a list of
    # modes that achieve that maximum.

    corrcoef_pairs = key_cc(score, profile, attribute_names, salience_flag)

    c_max_val_iter = (
        coefs[0] for (_, coefs) in corrcoef_pairs if coefs is not None
    )
    c_max_val = max(c_max_val_iter)
    keymode_attributes = [
        attr
        for (attr, coefs) in corrcoef_pairs
        if coefs is not None and coefs[0] == c_max_val
    ]
    return keymode_attributes



"""
Tests for amads/time/durdist2.py
"""

import math

from amads.algorithms.nnotes import nnotes
from amads.core.histogram import boundaries_to_centers, centers_to_boundaries
from amads.io.readscore import read_score
from amads.music import example
from amads.time.durdist1 import duration_distribution_1
from amads.time.durdist2 import duration_distribution_2


def assert_equal_dist1d(dist, correct, tol=1e-6):
    """Asserts that two 1d distributions are equal within a tolerance."""
    csum = 0.0
    dsum = 0.0
    for i in range(len(correct)):
        csum += correct[i]
        dsum += dist.data[i]
        assert (
            abs(dist.data[i] - correct[i]) < tol
        ), f"dist data mismatch at {i}: {dist.data[i]} vs {correct[i]}"
    assert abs(csum - 1) < tol, "'correct' does not sum to 1"
    assert abs(dsum - 1) < tol, "dist data does not sum to 1"


def assert_equal_dist2d(dist, correct, tol=1e-6):
    """Asserts that two 2d distributions are equal within a tolerance."""
    csum = 0.0
    dsum = 0.0
    for r in range(len(correct)):
        for c in range(len(correct[r])):
            csum += correct[r][c]
            dsum += dist.data[r][c]
            assert abs(dist.data[r][c] - correct[r][c]) < tol, (
                f"dist data mismatch at ({r},{c}): "
                + f"{dist.data[r][c]} vs {correct[r][c]}"
            )
    assert abs(csum - 1) < tol, "'correct' does not sum to 1"
    assert abs(dsum - 1) < tol, "dist data does not sum to 1"


def test_durdist_with_custom_bins():
    """
    Test duration_distribution_2 with custom bin centers.
    """
    # Load example MIDI file
    my_midi_file = example.fullpath("midi/sarabande.mid")
    assert my_midi_file is not None, "Example MIDI file not found."

    # Import MIDI using partitura
    myscore = read_score(my_midi_file, show=False)

    # Calculate duration distribution with custom bins
    centers = [0.05 * (2**i) for i in range(9)]  # Custom boundaries
    dd2 = duration_distribution_2(myscore, bin_centers=centers)

    print("Duration pair distribution, miditoolbox compatible:", dd2)
    print(dd2.data)
    correct = [
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.357, 0.047, 0.0, 0.013, 0.003, 0.0, 0.0],
        [0.0, 0.0, 0.050, 0.427, 0.023, 0.013, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.023, 0.017, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.013, 0.013, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ]
    assert_equal_dist2d(dd2, correct, 0.001)

    dd2 = duration_distribution_2(myscore, bin_centers=centers)
    print("Duration pair distribution:", dd2)
    print(dd2.data)
    correct = [
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.357, 0.047, 0.0, 0.013, 0.003, 0.0, 0.0],
        [0.0, 0.0, 0.05, 0.427, 0.023, 0.013, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.023, 0.017, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.013, 0.013, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ]
    assert_equal_dist2d(dd2, correct, 0.001)

    # test durdist1 too
    dd1 = duration_distribution_1(myscore, bin_centers=centers)
    print("Duration distribution 1, miditoolbox compatible:", dd1)
    print(dd1.data)
    correct = [0.0, 0.0, 0.4186, 0.5116, 0.0399, 0.0266, 0.0033, 0.0, 0.0]
    assert_equal_dist1d(dd1, correct, 0.001)

    dd3 = duration_distribution_1(
        myscore, bin_centers=centers, ignore_extrema=True
    )
    print("Duration distribution 1:", dd3)
    print(dd3.data)
    correct = [0.0, 0.4186, 0.5116, 0.0399, 0.0266, 0.0033, 0.0]
    assert_equal_dist1d(dd3, correct, 0.001)

    # see if we can relate the two results. First, reconstruct the counts from
    # the probabilities in dd2, using nnotes to get the number of notes
    n = nnotes(myscore)
    dc1 = [round(n * p) for p in dd1.data]
    n_pairs = n - 1  # number of note pairs
    dc2 = [[round(n_pairs * p) for p in row] for row in dd2.data]
    # derive duration counts by summing over rows
    dc2to1 = [0] * len(dc2)
    for col in range(len(dc2[0])):
        s = 0
        for row in range(len(dc2)):
            s += dc2[row][col]
        dc2to1[col] = s
    print("dc2 counts:", dc2)
    print("dc2 to dc1 counts:", dc2to1)
    print("dc1 counts:", dc1)

    # now, one of the duration counts in dc2to1 will be one less than in dc1,
    # because the first note has no "prev" note to form a pair with. This loop
    # will "fix" one and ony one count to make them match (or die trying)
    for i in range(len(dc1)):
        if dc1[i] == dc2to1[i] + 1:
            dc2to1[i] += 1
            break
    else:
        assert False, "Could not reconcile dd1 and dd2 counts"
    assert dc1 == dc2to1, "Duration counts from dd1 and dd2 do not match"

    # test durdist1 with default bins
    dd1_default = duration_distribution_1(myscore)
    print("Duration distribution 1 with default bins:", dd1_default)
    print(dd1_default.data)
    correct = [0.419, 0.0, 0.5116, 0.0, 0.0399, 0.0166, 0.010, 0.0033, 0.0]
    assert_equal_dist1d(dd1_default, correct, 0.001)


def test_long_dur():
    """when durations are outside the bin boundaries"""
    from amads.core.basics import Note, Part, Score

    score = Score()
    _ = Part(
        Note(onset=0.0, duration=20.0, pitch=60),
        Note(onset=21.0, duration=30.0, pitch=62),
        Note(onset=52.0, duration=40.0, pitch=64),
        parent=score,
    )

    # Calculate duration distribution
    dd2 = duration_distribution_2(score)

    print("Duration pair distribution with long durations:", dd2)
    for row in dd2.data:
        for val in row:
            assert (
                val == 0.0
            ), "Expected all zero distribution for long durations"

    dd2 = duration_distribution_2(
        score, bin_centers=boundaries_to_centers([0.3, 1.0, 3.0, 6.0], "log")
    )
    print(
        "Duration pair distribution with long durations, custom bins:", dd2.data
    )
    correct = [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]
    assert_equal_dist2d(dd2, correct)


def test_dd2_long_and_short():
    """
    Test duration_distribution_2 with long and short durations.
    """
    from amads.core.basics import Note, Part, Score

    score = Score()
    _ = Part(
        Note(onset=0.0, duration=0.01, pitch=60),  # very short
        Note(onset=0.02, duration=0.02, pitch=62),
        Note(onset=0.05, duration=8.0, pitch=64),  # very long
        Note(onset=8.1, duration=0.5, pitch=65),
        Note(onset=8.7, duration=2, pitch=66),
        Note(onset=11, duration=0.03, pitch=67),  # very short
        parent=score,
    )

    # Calculate duration distribution (default
    dd2 = duration_distribution_2(score)

    print("Duration pair distribution with long and short durations:", dd2)
    print(dd2.data)
    # Check that the very short and very long durations are handled correctly
    correct = [[0.0] * 9 for _ in range(9)]
    correct[3][7] = 1.0
    print('"correct" values', correct)
    assert_equal_dist2d(dd2, correct)

    # now test with explicit bin boundaries
    boundaries = [0.3, 1.0, 3.0, 6.0]
    dd2b = duration_distribution_2(
        score,
        bin_centers=boundaries_to_centers(boundaries, interpolation="log"),
    )
    print("Duration pair distribution with custom bins:", dd2b)
    print(dd2b.data)
    correct = [[0.2, 0.2, 0.2], [0.2, 0.0, 0.0], [0.2, 0.0, 0.0]]
    assert_equal_dist2d(dd2b, correct)


def test_centers_to_boundaries():
    """
    Test centers_to_boundaries function.
    """
    centers = [0.2, 0.5, 1.0, 2.0, 5.0]
    boundaries = centers_to_boundaries(centers, interpolation="log")
    print("Centers:", centers)
    print("Boundaries:", boundaries)
    correct = [
        math.sqrt(0.2 * 0.5),
        math.sqrt(0.5 * 1.0),
        math.sqrt(1.0 * 2.0),
        math.sqrt(2.0 * 5.0),
    ]
    for b, c in zip(boundaries, correct):
        assert abs(b - c) < 1e-6, f"Boundary mismatch: {b} vs {c}"

    boundaries = centers_to_boundaries(centers, interpolation="linear")
    print("Boundaries (linear):", boundaries)
    correct = [
        (0.2 + 0.5) / 2,
        (0.5 + 1.0) / 2,
        (1.0 + 2.0) / 2,
        (2.0 + 5.0) / 2,
    ]
    for b, c in zip(boundaries, correct):
        assert abs(b - c) < 1e-6, f"Boundary mismatch (linear): {b} vs {c}"



from typing import Optional, Sequence

import numpy as np

from amads.core.basics import Score

__author__ = "David Whyatt"

from amads.core.utils import key_num_to_name


class PolynomialContour:
    """A class for computing polynomial contour.

    As described in the FANTASTIC toolbox [1].
    This approach is discussed in detail in Müllensiefen and Wiggins (2011) [2].

    Polynomial Contour is constructed in 3 simple steps:

     - First, the onsets are first centred around the origin of the time axis,
       making a symmetry between the first onset and the last.

     - Then, a polynomial model is fit, seeking to predict the pitch values from
       a least squares regression of the centred onset times.

     - Finally, the best model is selected using Bayes' Information Criterion,
       stepwise and in a backwards direction.

    The final output is the coefficients of the first three non-constant terms,
    i.e. [c1, c2, c3] from p = c0 + c1t + c2t^2 + c3t^3.

    <small>**Author**: David Whyatt</small>

    Attributes
    ----------
    coefficients : list[float]
        The polynomial contour coefficients. Returns the first 3 non-constant
        coefficients [c1, c2, c3] of the final selected polynomial contour
        model. The constant term is not included as per the FANTASTIC
        toolbox specification.

    References
    ----------
     1. Müllensiefen, D. (2009). Fantastic: Feature ANalysis Technology
        Accessing STatistics (In a Corpus): Technical Report v1.5
     2. Müllensiefen, D., & Wiggins, G.A. (2011). Polynomial functions as a
        representation of melodic phrase contour.

    Examples
    --------
    Single note melodies return [0.0, 0.0, 0.0] since there is no contour:
    >>> pc = PolynomialContour(onsets=[1.0], pitches=[60])
    >>> pc.coefficients
    [0.0, 0.0, 0.0]

    Real melody examples:
    >>> test_pitches = [62, 64, 65, 67, 64, 60, 62]
    >>> test_case = Score.from_melody(pitches=test_pitches, durations=[1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0]) # duration
    >>> test_case_pc = PolynomialContour(test_case)
    >>> test_case_pc.onsets
    [0.0, 1.0, 2.0, 3.0, 4.0, 6.0, 7.0]

    >>> [round(x, 7) for x in test_case_pc.coefficients]  # Verified against FANTASTIC toolbox
    [-1.5014826, -0.2661533, 0.122057]

    The same result if this data comes from a score or directly.

    >>> test_onsets = [0.0, 1.0, 2.0, 3.0, 4.0, 6.0, 7.0]
    >>> test_case_pc.onsets == test_onsets
    True

    >>> test_pitches = [62, 64, 65, 67, 64, 60, 62]
    >>> test_2 = PolynomialContour(onsets=test_onsets, pitches=test_pitches)
    >>> test_2.onsets == test_onsets
    True

    >>> [round(x, 7) for x in test_2.coefficients]  # Verified against FANTASTIC toolbox
    [-1.5014826, -0.2661533, 0.122057]

    >>> twinkle = Score.from_melody([60, 60, 67, 67, 69, 69, 67, 65, 65, 64, 64, 62, 62, 60],
    ... [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0])
    >>> pc3 = PolynomialContour(twinkle)
    >>> [round(x, 7) for x in pc3.coefficients]  # Verified against FANTASTIC toolbox
    [-0.9535562, 0.2120971, 0.0]
    """

    def __init__(
        self,
        score: Optional[Score] = None,
        onsets: Optional[Sequence[float]] = None,
        pitches: Optional[Sequence[int]] = None,
    ):
        none_checks = (onsets is not None, pitches is not None)
        if any(none_checks) and not all(none_checks):
            raise ValueError(
                "onsets and pitches must be provided together, not one without the other."
            )

        if all(none_checks):
            if len(onsets) != len(pitches):
                raise ValueError(
                    f"onsets and pitches must have the same length, "
                    f"got {len(onsets)} and {len(pitches)}."
                )
            self.onsets = list(onsets)
            self.pitches = list(pitches)
        else:
            if score is None:
                raise ValueError(
                    "Provide either a Score or both onsets and pitches."
                )
            if not isinstance(score, Score):
                raise TypeError("Score should be a Score object.")

            self.onsets, self.pitches = self.get_onsets_and_pitches(score)

        self.coefficients = self.calculate_coefficients(
            self.onsets, self.pitches
        )

    def calculate_coefficients(
        self, onsets: list[float], pitches: list[int]
    ) -> list[float]:
        """Calculate polynomial contour coefficients for the melody.
        Main method for the PolynomialContour class.

        Parameters
        ----------
        onsets : list[float]
            List of onset times from the score
        pitches : list[int]
            List of pitch values from the score

        Returns
        -------
        list[float]
            First 3 coefficients [c1, c2, c3] of the polynomial contour, with zeros
            padded if needed. For melodies with fewer than 2 notes, returns [0.0, 0.0, 0.0]
            since there is no meaningful contour to analyze.
        """
        if len(onsets) <= 1:
            return [0.0, 0.0, 0.0]

        # Center onset times
        centered_onsets = self.center_onset_times(onsets)

        # Calculate polynomial degree
        m = len(onsets) // 2

        # Select best model using BIC
        return self.select_model(centered_onsets, pitches, m)

    def get_onsets_and_pitches(
        self, score: Score
    ) -> tuple[list[float], list[int]]:
        """Extract onset times and pitches from a Score object.

        Parameters
        ----------
        score : Score
            The Score object to extract data from

        Returns
        -------
        tuple[list[float], list[int]]
            A tuple containing (onset_times, pitch_values)
        """
        notes = score.get_sorted_notes()
        return [note.onset for note in notes], [note.key_num for note in notes]

    def center_onset_times(self, onsets: list[float]) -> list[float]:
        """Center onset times around their midpoint. This produces a symmetric axis
        of onset times, which is used later to fit the polynomial.

        For single-note melodies, returns [0.0] since there is no meaningful contour
        to analyze.

        Parameters
        ----------
        onsets : list[float]
            List of onset times to center

        Returns
        -------
        list[float]
            List of centered onset times. Returns [0.0] for single-note melodies.
        """
        if len(onsets) <= 1:
            return [0.0] * len(onsets)

        # Calculate midpoint using first and last onset times
        midpoint = (onsets[0] + onsets[-1]) / 2
        return [time - midpoint for time in onsets]

    def fit_polynomial(
        self, centered_onsets: list[float], pitches: list[int], m: int
    ) -> list[float]:
        """
        Fit a polynomial model to the melody contour using least squares regression.

        The polynomial has the form:
        p = c0 + c1*t + c2*t^2 + ... + cm*t^m

        where m = n // 2 (n = number of notes) and t are centered onset times.

        Parameters
        ----------
        centered_onsets : list[float]
            List of centered onset times
        pitches : list[int]
            List of pitch values
        m : int
            Maximum polynomial degree to use

        Returns
        -------
        list[float]
            The coefficients [c0, c1, ..., cm] of the fitted polynomial
        """

        n = len(pitches)
        if n <= 1:
            return [float(pitches[0]) if n == 1 else 0.0]

        # Create predictor matrix X where each column is t^i
        x = np.array(
            [[t**i for i in range(m + 1)] for t in centered_onsets], dtype=float
        )
        y = np.array(pitches, dtype=float)

        # Use numpy's least squares solver
        coeffs = np.linalg.lstsq(x, y, rcond=None)[0]

        return coeffs.tolist()

    def select_model(
        self, centered_onsets: list[float], pitches: list[int], m: int
    ) -> list[float]:
        """Select the best polynomial model using BIC in an exhaustive search
        over all subsets of polynomial terms.

        Tests all 2^(m+1) - 1 combinations of polynomial terms and selects
        the one with the best (lowest) BIC. The max degree is m = n // 2.

        Note: the search space grows as O(2^m).
        This is fine for shot melodies (up to c.30 notes, m <= 15).
        Longer melodies will be slow and need a review of this method for performance.

        Parameters
        ----------
        centered_onsets : list[float]
            List of centered onset times
        pitches : list[int]
            List of pitch values
        m : int
            Maximum polynomial degree to consider

        Returns
        -------
        list[float]
            The coefficients [c1, c2, c3] of the selected polynomial model,
            padded with zeros if the selected degree is less than 3.
        """
        max_degree = m
        pitches_array = np.array(pitches, dtype=float)
        x_full = np.array(
            [[t**i for i in range(max_degree + 1)] for t in centered_onsets]
        )

        # Start with maximum degree model
        best_fit = self.fit_polynomial(centered_onsets, pitches, m)
        # Pad to at least degree-3 so indexing [1],[2],[3] is always safe
        best_coeffs = np.zeros(max(max_degree + 1, 4))
        best_coeffs[: len(best_fit)] = best_fit
        best_bic = self._calculate_bic(
            best_coeffs[: max_degree + 1], x_full, pitches_array
        )

        for i in range(1, 2 ** (max_degree + 1)):
            binary = format(i, f"0{max_degree + 1}b")
            degrees = [j for j in range(1, max_degree + 1) if binary[j] == "1"]

            if not degrees:
                continue

            x = np.ones((len(centered_onsets), len(degrees) + 1))
            for j, degree in enumerate(degrees):
                x[:, j + 1] = [t**degree for t in centered_onsets]

            coeffs = np.linalg.lstsq(x, pitches_array, rcond=None)[0]

            # Build a full coefficient array (padded to at least degree 3)
            test_coeffs = np.zeros(max(max_degree + 1, 4))
            test_coeffs[0] = coeffs[0]
            for j, degree in enumerate(degrees):
                test_coeffs[degree] = coeffs[j + 1]

            bic = self._calculate_bic(
                test_coeffs[: max_degree + 1], x_full, pitches_array
            )

            if bic < best_bic:
                best_coeffs = test_coeffs
                best_bic = bic

        return [
            best_coeffs[1].item(),  # convert to native float
            best_coeffs[2].item(),
            best_coeffs[3].item(),
        ]

    def _calculate_bic(
        self, coeffs: np.ndarray, x: np.ndarray, y: np.ndarray
    ) -> float:
        """Calculate BIC for a set of coefficients.

        Emulates the FANTASTIC toolbox implementation, which uses stepAIC from
        the MASS package in R. Only non-zero coefficients are counted as
        parameters.

        If the max value is 0, then a small epsilon is added to RSS.
        We do this before taking the log to guard against
        the case of a perfect fit (RSS = 0 → log(0) = -inf).

        Parameters
        ----------
        coeffs : np.ndarray
            Coefficient array (length must match x.shape[1])
        x : np.ndarray
            Predictor matrix
        y : np.ndarray
            Response vector

        Returns
        -------
        float
            BIC value
        """
        predictions = np.dot(x, coeffs)
        residuals = predictions - y
        rss = np.sum(residuals**2)
        rss = max(rss, 1e-10)  # guard against log(0) on perfect fits
        n = len(y)

        # Count only non-zero coefficients as parameters
        n_params = np.sum(np.abs(coeffs) > 1e-10)

        return n * np.log(rss / n) + n_params * np.log(n)

    def plot(self, ax=None):
        """Plot the melody contour and the fitted polynomial curve.

        Displays pitch values at their centered onset times (scatter) with the
        selected polynomial fit overlaid (line). The y-axis is labelled with
        note names derived from the MIDI pitch numbers.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            Axes to draw on. If None, a new figure and axes are created
            with ``figsize=(8, 4)``.

        Returns
        -------
        matplotlib.axes.Axes
            The axes containing the plot, suitable for further customisation
            or embedding in a larger figure.

        Raises
        ------
        ImportError
            If matplotlib is not installed.

        """
        try:
            import matplotlib.pyplot as plt
            import matplotlib.ticker as ticker
        except ImportError as e:
            raise ImportError(
                "matplotlib is required for plotting. "
                "Install it with: pip install matplotlib"
            ) from e

        onsets = self.onsets
        pitches = self.pitches
        centered_onsets = self.center_onset_times(onsets)

        t = np.array(centered_onsets)
        c1, c2, c3 = self.coefficients

        # Recover the constant term c0 (not stored per FANTASTIC spec) as the
        # mean of the residuals after subtracting the known polynomial terms.
        poly_terms = c1 * t + c2 * t**2 + c3 * t**3
        c0 = float(np.mean(np.array(pitches, dtype=float) - poly_terms))

        t_smooth = np.linspace(t[0], t[-1], 300) if len(t) > 1 else t.copy()
        fit_curve = c0 + c1 * t_smooth + c2 * t_smooth**2 + c3 * t_smooth**3

        if ax is None:
            _, ax = plt.subplots(figsize=(8, 4))

        ax.scatter(
            centered_onsets,
            pitches,
            zorder=3,
            label="Notes",
            color="steelblue",
            s=60,
            linewidths=0.8,
            edgecolors="white",
        )
        ax.plot(
            t_smooth,
            fit_curve,
            color="tomato",
            linewidth=2,
            label=f"Contour [$c_{1}$={c1:.3f}, $c_{2}$={c2:.3f}, $c_{3}$={c3:.3f}]",
            zorder=2,
        )

        unique_pitches = sorted(set(pitches))
        ax.set_yticks(unique_pitches)
        ax.set_yticklabels(
            [key_num_to_name(p) for p in unique_pitches], fontsize=9
        )

        ax.xaxis.set_major_formatter(
            ticker.FuncFormatter(lambda x, _: f"{x:+.1f}")
        )
        ax.set_xlabel("Centered onset time", fontsize=10)
        ax.set_ylabel("Pitch", fontsize=10)
        ax.set_title("Polynomial contour", fontsize=11)
        ax.legend(fontsize=9)
        ax.grid(True, linestyle="--", alpha=0.4)
        ax.spines[["top", "right"]].set_visible(False)

        return ax



"""
Cross-correlations between pitch-class distributions and key profiles.


<small>**Author**: Tai Nakamura, Di Wang</small>


Reference
---------
https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=68 for more details
"""

__author__ = ["Tai Nakamura", "Di Wang"]

import math
from dataclasses import fields
from typing import List, Optional, Tuple

import numpy as np

import amads.pitch.key.profiles as prof
from amads.core.basics import Score
from amads.pitch.pcdist1 import pitch_class_distribution_1


def key_cc(
    score: Score,
    profile: prof.KeyProfile = prof.krumhansl_kessler,
    attribute_names: Optional[List[str]] = None,
    salience_flag: bool = False,
) -> List[Tuple[str, Optional[Tuple[float]]]]:
    """
    Calculate the correlation coefficients with specific pitch profiles.

    A score's pitch-class distribution is computed and generally,
    KeyProfiles come from existing data in profiles.py. Within each
    KeyProfile are one or more distributions, e.g. for "major" and
    "minor" keys, so you must specify which distributions you want
    correlations for.  Return a list of tuples, each containing the
    attribute name (e.g., "major") and the corresponding 12 correlation
    coefficients.

    When `salience_flag` is True, the pitch class distribution from the score
    (pcd) is replaced by a new one (pcd2) where each element is a weighted sum
    of the elements of pcd. The weights are rotated for each element. Thus,
    pcd2[i] = sum(pcd[j] * weight[(j + i) mod 12].

    The idea here is that the perception of significance of a certain pitch in
    a score depends not only on its naive unweighted frequency, but also (to a
    lesser extent) on the frequency of functionally harmonic pitches present
    in the score.

    Parameters
    ----------
    score: Score
        The score to analyze.

    profile: prof.KeyProfile
        The key profile to use for analysis.

    attribute_names: Optional[List[str]]
        List of attribute names that denote the particular PitchProfiles
        within the KeyProfile to compute correlations for. An example
        `attribute_names` for profile prof.vuvan could be
        `["natural_minor", "harmonic_minor"]`, which says to
        compute the cross-correlation between the pitch-class distribution
        of the score and both prof.vuvan's natural_minor and prof.vuvan's
        harmonic_minor. `None` can be supplied when we want to specify all
        valid pitch profiles within a given key profile.

    salience_flag: bool
        If True, apply salience pitch-wise bias weights to the score's
        pitch-class distribution.

    Returns
    -------
    List[Tuple[str, Optional[Tuple[float]]]]
        A list of tuples where each tuple contains the attribute name, from
        parameter `attribute_names`, and the corresponding 12-tuple of
        correlation coefficients. If an attribute name does not reference
        a valid data field within the specified key profile, it will yield
        `(`*attribute_name*`, None)`.

    Raises
    ------
    RuntimeError
        If the score or key profile contains equal pitch weights,
        resulting in correlation not being able to be computed.
    """

    # Get pitch-class distribution
    pcd = np.array([pitch_class_distribution_1(score, weighted=False).data])

    # Apply salience weighting if requested
    if salience_flag:
        # NOTE: this is not the weight vector,
        # the salience weights for the c-pitch in the pitch-class distribution
        # is [1, 0, 0.2, 0.17, 0.33, 0, 0, 0.5, 0, 0, 0.25, 0].
        # These weights form the first column of the 12x12 matrix salm.
        sal2 = [1, 0, 0.25, 0, 0, 0.5, 0, 0, 0.33, 0.17, 0.2, 0] * 2
        salm = np.zeros((12, 12))
        for i in range(salm.shape[0]):
            salm[i] = sal2[12 - i : 24 - i]
        pcd = np.matmul(pcd, salm.T)  # shape (1, 12)

    results = []

    true_attribute_names = attribute_names

    if true_attribute_names is None:
        true_attribute_names = [
            f.name
            for f in fields(profile)
            if f.name not in ["name", "literature", "about"]
        ]

    for attr_name in true_attribute_names:
        # ! we should probably treat the special attributes as proper attribute names
        if attr_name in ["name", "literature", "about"]:
            print(
                f"Warning! Attempting to access metadata in profile '{profile.name}"
            )
            results.append((attr_name, None))
            continue
        # Get the attribute from the profile
        attr_value = getattr(profile, attr_name, None)

        if attr_value is None:
            print(
                f"Warning: Attribute '{attr_name}' is invalid or None in profile '{profile.name}'"
            )
            results.append((attr_name, None))
            continue
        profiles_matrix = attr_value.as_canonical_matrix()
        correlations = tuple(_compute_correlations(pcd, profiles_matrix))
        if any(math.isnan(val) for val in correlations):
            raise RuntimeError(
                "key_cc has encountered either an invalid or equal weight"
                " score, or invalid pitch profile\n"
                f"correlations = {list(correlations)}\n"
                f"score pitch-class distribution = {list(pcd)}\n"
                f"profiles matrix = \n{profiles_matrix}\n"
            )
        results.append((attr_name, correlations))

    return results


def _compute_correlations(
    pcd: np.ndarray, profile_matrix: np.ndarray
) -> np.ndarray:
    """
    Compute correlations between pitch-class distribution and profile matrix.

    Parameters
    ----------
    pcd : np.ndarray
        Pitch-class distribution (1x12)
    profile_matrix : np.ndarray
        Profile matrix (Nx12 where N is number of keys/profiles)

    Returns
    -------
    np.ndarray
        Correlation coefficients
    """
    # Combine pcd and profile matrix for correlation computation
    combined_matrix = np.concatenate(
        (pcd, profile_matrix), axis=0
    )  # shape (25, 12)
    correlation_matrix = np.corrcoef(combined_matrix)

    # Extract correlations between pcd (first row) and all profiles
    correlations = correlation_matrix[0, 1:]

    return correlations



"""
Tests for basic normalization routines (norm module).
"""

import pytest

from amads.algorithms.norm import (
    euclidean_distance,
    manhattan_distance,
    normalize,
    shared_length,
)


def test_norm_raises():
    with pytest.raises(ValueError):
        normalize([0, 1, 2, 3], method="unrecognised string")


def test_shared_length_fails():
    profile_1 = [0, 1, 2, 3]
    profile_2 = [1, 2, 3, 4, 5]
    with pytest.raises(ValueError):
        shared_length(profile_1, profile_2)
    with pytest.raises(ValueError):
        manhattan_distance(profile_1, profile_2)
    with pytest.raises(ValueError):
        euclidean_distance(profile_1, profile_2)



"""
Test pitch list transformations

Functions for transforming pitch lists:
e.g., transposition, inversion, retrograde, rotation.

Most apply equally to any pitch class sequence,
some are more specific to tone rows.

Author: Mark Gotham, 2021

License: Creative Commons Attribution-ShareAlike 4.0 International License
https://creativecommons.org/licenses/by-sa/4.0/

Source: Gotham and Yust, Serial Analyser, DLfM 2021
https://github.com/MarkGotham/Serial_Analyser
"""

from amads.pitch.serial import pair_swap_krenek, rotate_hexachords
from amads.pitch.transformations import (
    invert,
    pitches_to_intervals,
    retrograde,
    rotate,
    transpose_by,
    transpose_to,
)


def test_transpose():
    row_boulez = [3, 2, 9, 8, 7, 6, 4, 1, 0, 10, 5, 11]
    zero_boulez = transpose_to(row_boulez, start=0)
    assert zero_boulez == [0, 11, 6, 5, 4, 3, 1, 10, 9, 7, 2, 8]
    trans_boulez = transpose_to(zero_boulez, start=3)
    assert trans_boulez == [3, 2, 9, 8, 7, 6, 4, 1, 0, 10, 5, 11]
    trans_by_boulez = transpose_by(trans_boulez, semitones=2)
    assert trans_by_boulez == [5, 4, 11, 10, 9, 8, 6, 3, 2, 0, 7, 1]


def test_rotate():
    luto = [0, 6, 5, 11, 10, 4, 3, 9, 8, 2, 1, 7]
    for i in range(12):
        row = rotate(luto, i)
        assert row[0] == luto[i]


def test_invert():
    test_set = [0, 1, 4, 6]
    assert invert(test_set) == [0, 11, 8, 6]


def test_pitches_to_intervals():
    test_row_up = [x for x in range(12)]
    assert pitches_to_intervals(test_row_up) == [1] * 11
    test_row_down = test_row_up[::-1]
    assert pitches_to_intervals(test_row_down) == [11] * 11


def test_rotate_hexachords():
    """Using Krenek's example"""
    row_krenek = [5, 7, 9, 10, 1, 3, 11, 0, 2, 4, 6, 8]
    rotated_krenek = rotate_hexachords(row_krenek)
    assert len(rotated_krenek) == 7
    assert rotated_krenek[-1] == row_krenek


def test_pair_swap_and_retrograde():
    """Using Krenek's example"""
    test_row_ = [9, 2, 3, 6, 5, 1, 7, 4, 8, 0, 10, 11]
    test_pair_swap_krenek = pair_swap_krenek(test_row_)
    assert len(test_pair_swap_krenek) == 13
    assert test_pair_swap_krenek[-1] == retrograde(test_row_)
